/* ------------------------------------------------------------------------
System       : Manx Aztec C65 Version 3.2b
               MS-DOS cross-development environment
Platform     : Apple IIe 128K PRODOS 8
Program      : dpshow(C) Copyright Bill Buckels 2013.
               All rights reserved.

Description  : Apple //e ProDOS 8 DHGR Slideshow Program

			   DHGR formats supported:  A2FC, BIN AUX pairs, and
			                            DHX and DHR xpacked images.

               Uses Page 2 Double Hi-Res to avoid
               ProDOS Sys Program Load Address at 0x2000

Written by   : Bill Buckels
Date Written : February 2013

Revision     : 1.0 First Release
Licence      : You may use this program for whatever you wish as long
               as you agree that Bill Buckels has no warranty or
               liability obligations whatsoever from said use.
------------------------------------------------------------------------ */

#include <g3.h>
#include <stdio.h>
#include <prodir.h>

#include "paclist.inc"

char *title = "dpshow";
char *copyright = "(C) Copyright Bill Buckels 2013.All rights reserved.";

int mystuff()
{
	FILE *fp;
	int cnt = 0, idx, valid = 1;
	char *buf = (char *)4192;

	if((fp=fopen("copyread","r"))==NULL) return cnt;


	/* allow 20 lines for copyright and usage */
	for (cnt = 0; cnt < 20;) {
		buf[0] = 0;
		if (fgets(buf,80,fp) == -1) break;

        /* signature line must match */
		if (cnt == 0) {
		    for (idx = 0;copyright[idx] != 0; idx++) {
				if (copyright[idx] != buf[idx]) {
				   valid = 0;
				   break;
				}
			}
			if (valid == 0) break; /* outa' here */
			scr_curs(cnt,0);puts(title);cnt++;
			scr_curs(cnt,0);puts(copyright);cnt++;
		}
		else {
			for (idx = 0;buf[idx] != 0;) {
				if (buf[idx] == 13||buf[idx] == 10||buf[idx]==0) buf[idx] = 0;
				else {
					if (buf[idx] < 32 || buf[idx] > 122) {
						valid = 0;
						break;
					}
					idx++;
				}
			}
		    if (valid == 0) break; /* outa' here */
			scr_curs(cnt,0);puts(buf);cnt++;
		}
	}
	fclose(fp);
	scr_curs(cnt,0);
	return cnt;

}

/* the following is what I eventually ended-up with as a reasonable
   compromise for the structure of a simple slideshow program for
   ProDOS using a picture file list. It is a little more elegant than
   some of the other simple slideshow programs that I wrote in this
   same environment over 20 years back. I have done my best not to stop
   the show even if some of the pictures are missing or in an unsupported
   format. */

main()
{
	FILE *fp;
	unsigned filesize;
	int idx,status,started=0,key=0,c,cnt;
	char picname[82];

	scr_clear();
	scr_curs(0,0);

	/* set page 2 double hires - we need to use page 2 of course because
	   ProDOS SYS programs load at the same address as the page 1 screen */
	d2hireson();

	/* clear screen by copying memory hole in sys program
	to page2 hi-res screen in auxiliary memory */
	maintoaux(0x4000,0x4000+8191,0x4000);

    /* loop through the slideshow and restart at the beginning until the
       user presses the escape key... */

    /* on a clear display you can C for(;;) (forever)

       At my age I prefer unconditional loops with clean breaks and a few flags
       since the language supports it and I find it somewhat more readable realizing
       of course that a break amounts to no more than a goto anyway.

       And you guys who prefer while(1==0) can have it! That's a condition that
       I'd prefer not to imagine. The highway, like the Apple II and this slideshow,
       goes on forever. presumably. */

	for (;;) {

		status = -1;
		if((fp=fopen("PACLIST","r"))==NULL) {
			/* first time through try to create piclist if it does not exist */
			if (started == 0) {
			   fp = makepaclist();
			}

			if (fp == NULL) break;
		}
		started = 1;
		picname[0] = 0; /* gigo - blank the line before calling fgets */
		/* get the first filename */
		if (fgets(picname,80,fp) == -1) {
			fclose(fp);
			break;
		}
		/* if garbage on the first line outa' here */
		if (picname[0] < 33 || picname[0] > 122) {
			fclose(fp);
			break;
		}

		/* clear keyboard */
		while (kbhit()!=0)clearkey();
		status = cnt = 0;

        /* before I allow an image to load, I have established a rather lengthy
           interview process... basically for the hellery of it and to satisfy the
           arguably sadistic wish to deny entry unless ones' details are in in order
           that dwells in the heart of every good gatekeeper and passes for fun in
           the soul of every good programmer. */

        /* if the first character on the line has an ascii value of a space character or below
           this program will fail... I do not support blank lines and text files must
           be in Apple Format or Unix format... an MS-DOS paired 13,10 EOL will cause
           fgets to read a blank line after each valid line so only the first line will
           be read and if it is an image entry the first image will load repetively and
           frankly Scarlet, I don't give a damn. My main concern here is reading past
           eof so my main concern is keeping the garbage out of here or there is a potential
           for a less-than spectacular crash possibly with the read loop going into a
           tailspin first... that is, if we don't have a garbage filter and some
           belt and suspender rules. a more cavalier approach could be applied here
           with the potential for disasterous results bur I also could have applied
           exhaustive format validation... since I am not Jack Crenshaw working on the
           Apollo Mission I am confident no lives will be lost in space with what I put here. */

		/* line validation */
		while(picname[0] > 32 && picname[0] < 123) {

			/* preload integrity check */
			picname[64] = 0; /* truncate the line to the maximum length
			                    of a ProDOS Pathname */
			idx=0;  /* replace carriage return with null, ignore lines beginning
			           with invalid characters etc. */

			while(picname[idx]!=0)
			{
				 /* if I find an ascii value below the prefix character value (46)
				    or any above a lower-case z I truncate the line there...
				    comments can be on the line separated from the picname
				    by an ascii value below 46 (a space or a pound sign are ok)
				    but the line must not exceed 80 characters... I had to draw the
				    line somewhere... */
				 if(picname[idx] < 46 || picname[idx] > 122)picname[idx]= 0;
				 else idx++;
			}

            /* comments can also be on a separate line. the rule here is almost
               the same as noted above for trailing comments on the picture entry line
               except that the first character on the line must be an ascii value
               from 33 to 45 (spaces are not allowed, pound signs work just fine) */

			if (picname[0] != 0) {

				/* preload keycheck - skip loading the image if
				   the user leans on the keyboard and try to avoid
				   partially loading raster based images because they
				   can look like heck */
				key = kbhit();
				if (key!=0) {
					clearkey();
					cnt++;
				}
				else {
					c = d2pld(picname);
					if (c< 0) c = d2lode(picname);
					if (c > -1) cnt++;

					/* if a load error skip this slide, otherwise... */
					/* advance on timeout or advance or end on keypress */
					if (c==0) key = wait(10,XTIME,YTIME);

				} /* nobody leaned on the keyboard to advance to the next slide */
			} /* the format for the picture name entry in the piclist is valid */
			else {
				/* picname[0] == 0 - truncated line - began with garbage */
				key = kbhit();
				if (key!=0)clearkey();

			}
			/* if escape key is pressed end slideshow */
			if (key == 27) break;
			picname[0] = 0; /* gigo - make sure no garbage is going in to fgets
			                          trust nothing and no-one */
			if (fgets(picname,80,fp) == -1)break;
		}
		fclose(fp);

		/* if all the pictures are missing or invalid or if
		   the entire piclist is not correctly formatted
		   the count will be zero...
		   I report this as an error before the program exits.
		   I don't bother reporting the occasional missing picture
		   or one that didn't load vs. others that did...

		   I'll leave that for the user to figure-out.
		   I figured-out that enough is enough already so let them
		   give their brains some exercise if they even notice
		   some stuff not happening... empathy is no match for stupidity
		   anyway and a good programmer like a good shepherd knows
		   that it is pointless to guard a dead sheep, even a brain-dead one.

		   */

		if (cnt == 0)status = -1;
		if (key == 27 || cnt == 0)break;

		/* bottom of loop - start over at the beginning of
		   the slideshow piclist...
		   on a clear display you can C for(;;) (forever)
		   or is it on a clear disk you can seek forever? */

	}

    /* either the user has pressed the escape key to break our magic spell
       or the piclist is missing, or all the pictures are gone...
       but now it's time to say goodbye to all our company... */
	d2flood(0,0,139,191,0);
	d2hiresoff();
	scr_clear();
	scr_curs(0,0);

	/* clear keyboard */
	while (kbhit()!=0)clearkey();

	if (status!=0) {
		puts("!PACLIST missing, invalid, or could not be opened or created!");
		scr_curs(1,0);
		puts("Press Any Key...");
		getch();
		scr_clear();
		scr_curs(0,0);
  	}

  	/* display the copyright if one exists
  	   but if they just copy the program or remove it
  	   let them go ahead... there is only fame to be had from this old stuff
  	   anyway and certainly not fortune. enough crippleware exists in the world
  	   without me creating more. I am just happy to finish a program after
  	   all these years...
  	   (jk of course, but I am delighted that it's not crashing... so far:)

  	   hope you enjoyed the read even if the code is only k&r and I trust that
  	   something proved interesting or informative.

  	   for much much more, read the G2 library code and review some of
  	   my other fun little Aztec C (and other) projects.

  	   Bill Buckels
  	   February 4, 2013

  	   */
 	if (mystuff() != 0) {
		puts("\nPress Any Key...");
		getch();
		scr_clear();
		scr_curs(0,0);
	}

	_exit(); /* Aztec C exits with a warm boot back to ProDOS
	            if everything has gone well... if one crashes into the
	            monitor there is still work to do. */
}

