/* ------------------------------------------------------------------------ */
/* BMP2LO.C (C) Copyright Bill Buckels 2009, 2012, 2013                     */
/* All Rights Reserved.                                                     */
/*                                                                          */
/* Licence Agreement                                                        */
/* -----------------                                                        */
/*                                                                          */
/* You have a royalty-free right to use, modify, reproduce and              */
/* distribute this source code in any way you find useful,                  */
/* provided that you agree that Bill Buckels has no warranty obligations    */
/* or liability resulting from said distribution in any way whatsoever.     */
/* If you don't agree, remove this source code from your computer now.      */
/*                                                                          */
/* Written by   : Bill Buckels                                              */
/* Email:         bbuckels@escape.ca                                        */
/*                                                                          */
/* Purpose      : This utility will allow you to convert to 				*/
/*                Apple II Double Lo-res 80 x 48 x 16 color images		    */
/*                                                                          */
/*                from IBM-PC graphics files in the following formats:		*/
/*																			*/
/*                CGA 320 x 200 x 4 color BASIC BSAVED IMAGE (.BAS) Files 	*/
/*                CGA 320 x 200 x 4 color ZSOFT .PCX Files					*/
/*				  EGA 16 color Windows .BMP Files up to 320 x 200			*/
/*																			*/
/*                24 - Bit .BMP files in several - sizes                    */
/*				  When converting 24 - bit BMP files Single Lo-Res and		*/
/*                Double Lo-Res files can be created.                       */
/*																			*/
/* Revision     : 1.0 First Release                                         */
/*                2.0 Added support for ramfont for newer computers.        */
/*                3.0 Fixed Auxiliary Colors                                */
/*                    Added Support for 24 bit "canvas-style" bmp's         */
/* ------------------------------------------------------------------------ */
/* Written in Large Model 16 bit Microsoft C (MSC) Version 8.00c            */
/* Note: Run in an MS-DOS emulator like DOSBox if you can't run it raw.     */
/* ------------------------------------------------------------------------ */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <dos.h>
#include <bios.h>
#include <io.h>
#include <malloc.h>
#include <conio.h>

#include "ramfont.h"

/* ------------------------------------------------------------------------ */
/* Declarations, Vars. etc.                                                 */
/* ------------------------------------------------------------------------ */

/* Bitmap Header structures */
typedef struct tagRGBQUAD
{
    unsigned char    rgbBlue;
    unsigned char    rgbGreen;
    unsigned char    rgbRed;
    unsigned char    rgbReserved;
} RGBQUAD;

typedef struct tagBITMAPINFOHEADER
{
    unsigned long   biSize;
    unsigned long   biWidth;
    unsigned long   biHeight;
    unsigned int    biPlanes;
    unsigned int    biBitCount;
    unsigned long   biCompression;
    unsigned long   biSizeImage;
    unsigned long   biXPelsPerMeter;
    unsigned long   biYPelsPerMeter;
    unsigned long   biClrUsed;
    unsigned long   biClrImportant;
} BITMAPINFOHEADER;

/* constant for the biCompression field */
#define BI_RGB      0L

typedef struct tagBITMAPFILEHEADER
{
    unsigned char   bfType[2];
    unsigned long   bfSize;
    unsigned int    bfReserved1;
    unsigned int    bfReserved2;
    unsigned long   bfOffBits;
} BITMAPFILEHEADER;

/* static structures for processing input file */
BITMAPFILEHEADER BitMapFileHeader;
BITMAPINFOHEADER bmi;

unsigned char *szTitle[]= {
"                                   ",
" BMP2LO(C) Version 3.0             ",
" Copyright Bill Buckels 2013       ",
" All Rights Reserved.              ",
" Distributed as FreeWare.          ",
" Email: bbuckels@escape.ca         ",
"                                   ",
" Press Any Key To Continue...      ",
" --------------------------------- ",
" To Position Clip Box - Arrow Keys ",
" Page Up, Page Down, Home, and End ",
"                                   ",
" 'S' or [ENTER] to  Save           ",
" 'Q' or [ESC] to Quit              ",
" 'H' For Help                      ",
"                                   ",
" 0-9 Toggle Color (Numeric Keys)   ",
" A-F Toggle Color (Alpha Keys)     ",
"                                   ",
NULL};

unsigned char *szTextTitle =
    "BMP2LO(C) Version 3.0 Copyright Bill Buckels 2013\n"
    "All Rights Reserved.";

#define TRUE     1
#define FALSE    0
#define SUCCESS  0
#define VALID    SUCCESS
#define FAILURE  -1
#define INVALID  FAILURE

#define MCGA '\x13'
#define TEXT '\x03'

#define ASCIIZ     '\x00'
#define CRETURN    '\r'
#define LFEED      '\n'

#define ENTERKEY   '\x0d' /* character generated by the Enter Key          */
#define ESCKEY     '\x1b' /* character generated by the Esc key            */
#define FUNCKEY    '\x00' /* first character generated by function keys    */
#define UPARROW    'H'    /* second character generated by up-arrow key    */
#define DOWNARROW  'P'    /* second character generated by down-arrow key  */
#define LTARROW    'K'    /* second character generated by left-arrow key  */
#define RTARROW    'M'    /* second character generated by right-arrow key */
#define PGUP       'I'    /* second character generated by page up key     */
#define PGDOWN     'Q'    /* second character generated by page down key   */
#define HOMEKEY    'G'    /* second character generated by home key        */
#define ENDKEY     'O'    /* second character generated by end key         */

/* second character generated by numerical fkeys */
/* starting at character 59                      */
/* ending at character 68                        */
#define F1         ';'
#define F2         '<'
#define F3         '='
#define F4         '>'
#define F5         '?'
#define F6         '@'
#define F7         'A'
#define F8         'B'
#define F9         'C'
#define F10        'D'

#define FRAMESIZE ((unsigned)64000)
#define FRAMEADDR ((unsigned char *)0xa0000000l)

/* middle of screen */
#define XMIN 0
#define YMIN 0
#define XMOS 159
#define YMOS 99
#define XMAX 319
#define YMAX 199

#define SCREENWIDTH  (XMAX + 1)
#define RASTERWIDTH  SCREENWIDTH
#define SCREENHEIGHT (YMAX + 1)
#define CELL_SIZE    8


// some "helpful" macros

#define vload() memcpy(FRAMEADDR,(unsigned char *)&rawbuffer[0],FRAMESIZE)
#define vsave() memcpy((unsigned char *)&rawbuffer[0],FRAMEADDR,FRAMESIZE)
#define zap(x) memset(FRAMEADDR,x,FRAMESIZE)
#define getpixel(x,y) rawbuffer[(y*RASTERWIDTH)+x]


// function prototypes

unsigned char GetMcgaPaletteIndex(unsigned int cgacolor),
      lsb(unsigned int),
      msb(unsigned int),
      SetCrtMode(unsigned char);

unsigned int byteword(unsigned char, unsigned char);

int BMP16_Read(unsigned char *),
    BSAVE_Read(unsigned char *),
    CheckForCGAPCX(unsigned char *),
    EatKeys(),
    GetVGAIndex(unsigned char, unsigned char, unsigned char),
    LoadPalette(void),
    MakeBMP(unsigned char *, unsigned char *),
    PCX_Read(unsigned char *),
    savelofragment(unsigned char *, int, int);

void LineBox(int, int, int, int, unsigned int),
     PCMidFont(unsigned char *, int, int, int, int, int),
     PCRamFont(unsigned char *, int, int, int, int, int),
     PutPixel(int, int, unsigned int, unsigned char *),
     setlopixel(unsigned char,int, int, int),
     SetPalette(),
     ShowTitle(void),
     TogglePalette(unsigned char),
     XBox(int, int, int, int),
     XPixel(int, int, unsigned char *);

int appletop = FALSE, lores = FALSE;

unsigned char outlinecolor;
unsigned char drawcolor;
unsigned char *rawbuffer;


#define NUM_MCGA_COLORS    256
#define NUM_RGB_COLORS     3

unsigned char rgbinfo[NUM_MCGA_COLORS][NUM_RGB_COLORS];  /* the vga palette */

enum {  BLACK = 0,
        BLUE,
        GREEN,
        CYAN,
        RED,
        MAGENTA,
        BROWN,
        GRAY,
        WHITE,
        LBLUE,
        LGREEN,
        LCYAN,
        LRED,
        LMAGENTA,
        YELLOW,
        BWHITE,
        NUM_VGA_COLORS};


/* the following 4 palettes are used to troll
   for standard colors in the order given below */

/* 16 Color BMP style palette (probably PBRUSH from Windows 3.1) */
unsigned char rgbBmpArray[NUM_VGA_COLORS][NUM_RGB_COLORS]={
0x00, 0x00, 0x00,    // BLACK
0x00, 0x00, 0xBF,    // BLUE
0x00, 0xBF, 0x00,    // GREEN
0x00, 0xBF, 0xBF,    // CYAN
0xBF, 0x00, 0x00,    // RED
0xBF, 0x00, 0xBF,    // MAGENTA
0xBF, 0xBF, 0x00,    // BROWN
0x80, 0x80, 0x80,    // GRAY
0xC0, 0xC0, 0xC0,    // WHITE
0x00, 0x00, 0xFF,    // LBLUE
0x00, 0xFF, 0x00,    // LGREEN
0x00, 0xFF, 0xFF,    // LCYAN
0xFF, 0x00, 0x00,    // LRED
0xFF, 0x00, 0xFF,    // LMAGENTA
0xFF, 0xFF, 0x00,    // YELLOW
0xFF, 0xFF, 0xFF};   // BWHITE

/* 16 Color BMP style palette (Windows XP) */
/* notice that these johnny-come-latelies reversed GRAY and WHITE */
unsigned char rgbXmpArray[NUM_VGA_COLORS][NUM_RGB_COLORS]={
0x00, 0x00, 0x00,  	 // BLACK
0x00, 0x00, 0x80,  	 // BLUE
0x00, 0x80, 0x00,  	 // GREEN
0x00, 0x80, 0x80,  	 // CYAN
0x80, 0x00, 0x00,  	 // RED
0x80, 0x00, 0x80,  	 // MAGENTA
0x80, 0x80, 0x00,  	 // BROWN
0x80, 0x80, 0x80,  	 // GRAY
0xC0, 0xC0, 0xC0,  	 // WHITE
0x00, 0x00, 0xFF,  	 // LBLUE
0x00, 0xFF, 0x00,  	 // LGREEN
0x00, 0xFF, 0xFF,  	 // LCYAN
0xFF, 0x00, 0x00,  	 // LRED
0xFF, 0x00, 0xFF,  	 // LMAGENTA
0xFF, 0xFF, 0x00,  	 // YELLOW
0xFF, 0xFF, 0xFF};	 // BWHITE

// VGA Palette
unsigned char rgbVgaArray[NUM_VGA_COLORS][NUM_RGB_COLORS]={
0x00, 0x00, 0x00,            /* BLACK    */
0x00, 0x00, 0xFF,            /* BLUE     */
0x00, 0xFF, 0x00,            /* GREEN    */
0x00, 0xFF, 0xFF,            /* CYAN     */
0xFF, 0x00, 0x00,            /* RED      */
0xFF, 0x00, 0xFF,            /* MAGENTA  */
0xFF, 0xFF, 0x00,            /* BROWN    */
0x55, 0x55, 0x55,            /* GRAY     */
0xC0, 0xC0, 0xC0,            /* WHITE    */
0x55, 0x55, 0xFF,            /* LBLUE    */
0x55, 0xFF, 0x55,            /* LGREEN   */
0x55, 0xFF, 0xFF,            /* LCYAN    */
0xFF, 0x55, 0x55,            /* LRED     */
0xFF, 0x55, 0xFF,            /* LMAGENTA */
0xFF, 0xFF, 0x55,            /* YELLOW   */
0xFF, 0xFF, 0xFF};           /* BWHITE   */

// 16 color ZSOFT PCPAINT PCX style palette
unsigned char rgbPcxArray[NUM_VGA_COLORS][NUM_RGB_COLORS]={
0x00, 0x00, 0x00,            /* BLACK    */
0x00, 0x00, 0xAA,            /* BLUE     */
0x00, 0xAA, 0x00,            /* GREEN    */
0x00, 0xAA, 0xAA,            /* CYAN     */
0xAA, 0x00, 0x00,            /* RED      */
0xAA, 0x00, 0xAA,            /* MAGENTA  */
0xAA, 0xAA, 0x00,            /* BROWN    */
0x55, 0x55, 0x55,            /* GRAY     */
0xAA, 0xAA, 0xAA,            /* WHITE    */
0x55, 0x55, 0xFF,            /* LBLUE    */
0x55, 0xFF, 0x55,            /* LGREEN   */
0x55, 0xFF, 0xFF,            /* LCYAN    */
0xFF, 0x55, 0x55,            /* LRED     */
0xFF, 0x55, 0xFF,            /* LMAGENTA */
0xFF, 0xFF, 0x55,            /* YELLOW   */
0xFF, 0xFF, 0xFF};           /* BWHITE   */


/* this palette is in the actual apple II lores color order */
/* this shows us approximately what we will end-up with */
unsigned char rgbLoresArray[NUM_VGA_COLORS][NUM_RGB_COLORS]={
  0,   0,   0,	/* black    */
227,  30,  96,	/* red      */
 96,  78, 189,	/* dk blue  */
255,  68, 253,	/* purple   */
  0, 163,  96,	/* dk green */
156, 156, 156,	/* gray     */
 20, 207, 253,	/* med blue */
208, 195, 255,	/* lt blue  */
 96, 114,   3,	/* brown    */
255, 106,  60, 	/* orange   */
156, 156, 156,	/* grey     */
255, 160, 208,	/* pink     */
 20, 245,  60,	/* lt green */
208, 221, 141,	/* yellow   */
114, 255, 208, 	/* aqua     */
255, 255, 255};	/* white    */

#define LOBLACK   	0
#define LORED     	1
#define LODKBLUE 	2
#define LOPURPLE  	3
#define LODKGREEN	4
#define LOGRAY    	5
#define LOMEDBLUE	6
#define LOLTBLUE 	7
#define LOBROWN   	8
#define LOORANGE  	9
#define LOGREY    	10
#define LOPINK    	11
#define LOLTGREEN	12
#define LOYELLOW  	13
#define LOAQUA    	14
#define LOWHITE   	15

unsigned char VGAtoApple[NUM_VGA_COLORS]={
	LOBLACK,
	LODKBLUE,
	LODKGREEN,
	LOMEDBLUE,
	LORED,
	LOPURPLE,
	LOBROWN,
	LOGRAY,
	LOWHITE,
	LOLTBLUE,
	LOLTGREEN,
	LOAQUA,
	LOORANGE,
	LOPINK,
	LOYELLOW,
	LOWHITE};

// to rollback the colors in case we can't automatically assign
unsigned char VGAtoAppleSaved[NUM_VGA_COLORS]={
	LOBLACK,
	LODKBLUE,
	LODKGREEN,
	LOMEDBLUE,
	LORED,
	LOPURPLE,
	LOBROWN,
	LOGRAY,
	LOWHITE,
	LOLTBLUE,
	LOLTGREEN,
	LOAQUA,
	LOORANGE,
	LOPINK,
	LOYELLOW,
	LOWHITE};


unsigned char awinoldcolors[16][4] = {
	0,0,0,      LOBLACK,    /* black    */
	208,0,48,	LORED,      /* red      */
	0,0,128,	LODKBLUE,   /* dk blue  */
	255,0,255,	LOPURPLE,   /* purple   */
	0,128,0,	LODKGREEN,  /* dk green */
	128,128,128,LOGRAY,     /* gray     */
	0,0,255,	LOMEDBLUE,  /* med blue */
	96,160,255,	LOLTBLUE,   /* lt blue  */
	128,80,0,	LOBROWN,    /* brown    */
    255,128,0,	LOORANGE,   /* orange   */
    192,192,192,LOGREY,     /* grey     */
    255,144,128,LOPINK,     /* pink     */
    0,255,0,	LOLTGREEN,  /* lt green */
    255,255,0,	LOYELLOW,   /* yellow   */
    64,255,144,	LOAQUA,     /* aqua     */
    255,255,255,LOWHITE};   /* white    */

unsigned char awinnewcolors[16][4] = {
	0,0,0,      LOBLACK,    /* black    */
	157,9,102,  LORED,      /* red      */
	42,42,229,  LODKBLUE,   /* dk blue  */
	199,52,255, LOPURPLE,   /* purple   */
	0,118,26,   LODKGREEN,  /* dk green */
	128,128,128,LOGRAY,     /* gray     */
	13,161,255, LOMEDBLUE,  /* med blue */
	170,170,255,LOLTBLUE,   /* lt blue  */
	85,85,0,    LOBROWN,    /* brown    */
    242,94,0,   LOORANGE,   /* orange   */
    192,192,192,LOGREY,     /* grey     */
    255,137,229,LOPINK,     /* pink     */
    56,203,0,   LOLTGREEN,  /* lt green */
    213,213,26, LOYELLOW,   /* yellow   */
    98,246,153, LOAQUA,     /* aqua     */
    255,255,254,LOWHITE};   /* white    */

/* the following is used to remap
   double lo res 4 bit colors
   from bank 0 to bank 1 */
unsigned char dloauxcolor[16] = {
	0,8,1,9,2,10,3,11,4,12,5,13,6,14,7,15};

/* the following is used to remap
   double lo res 4 bit colors
   from bank 1 to bank 0 */
unsigned char dlomaincolor[16] = {
	0,2,4,6,8,10,12,14,1,3,5,7,9,11,13,15};

/* our working copy of the apple II lores colors */
/* this is in Apple II order */
unsigned char rgbArray[NUM_VGA_COLORS][NUM_RGB_COLORS]={
  0,   0,   0,	/* black    */
227,  30,  96,	/* red      */
 96,  78, 189,	/* dk blue  */
255,  68, 253,	/* purple   */
  0, 163,  96,	/* dk green */
156, 156, 156,	/* gray     */
 20, 207, 253,	/* med blue */
208, 195, 255,	/* lt blue  */
 96, 114,   3,	/* brown    */
255, 106,  60, 	/* orange   */
156, 156, 156,	/* grey     */
255, 160, 208,	/* pink     */
 20, 245,  60,	/* lt green */
208, 221, 141,	/* yellow   */
114, 255, 208, 	/* aqua     */
255, 255, 255};	/* white    */


/* 16 Color VGA style palettes with RGB values
   from corresponding Apple II lores colors */
unsigned char rgbOriginalArray[NUM_VGA_COLORS][NUM_RGB_COLORS]={
  0,   0,   0,    // BLACK    - black
 96,  78, 189,    // BLUE     - dk blue
  0, 163,  96,    // GREEN    - dk green
 20, 207, 253,    // CYAN     - med blue
227,  30,  96,    // RED      - red
255,  68, 253,    // MAGENTA  - purple
 96, 114,   3,    // BROWN    - brown
156, 156, 156,    // GRAY     - gray or grey
255, 255, 255,    // WHITE    - white
208, 195, 255,    // LBLUE    - lt blue
 20, 245,  60,    // LGREEN   - lt green
114, 255, 208,    // LCYAN    - aqua
255, 106,  60,    // LRED     - orange
255, 160, 208,    // LMAGENTA - pink
208, 221, 141,    // YELLOW   - yellow
255, 255, 255};   // BWHITE   - white

enum {  CGA_BLACK = 0,
        CGA_CYAN,
        CGA_MAGENTA,
        CGA_WHITE,
        NUM_CGA_COLORS};

/* a microsoft compatible bsaved image format descriptor */
unsigned char BSAVED_header[7]={
    '\xfd','\x00','\xb8','\x00','\x00','\x00','\x40'};


/* bmiHeader.biClrUsed and bmiHeader.biClrImportant fields
   will vary depending on the process that created the bmp
   but the other fields (below) should be invariant. */
unsigned char BMP_checkheader[] ={
0x42, 0x4D, 0x76, 0x7D, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x76, 0x00, 0x00, 0x00, 0x28, 0x00,
0x00, 0x00, 0x40, 0x01, 0x00, 0x00, 0xC8, 0x00,
0x00, 0x00, 0x01, 0x00, 0x04, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x7D, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00};


/* ------------------------------------------------------------------------ */
/* Low Level Video Routines, drawing routines, etc.                         */
/* ------------------------------------------------------------------------ */

unsigned char SetCrtMode(unsigned char vidmode)
{
  union REGS inregs, outregs;

  /* set mode */
  inregs.h.ah = 0;
  inregs.h.al = vidmode;
  int86(0x10, &inregs, &outregs);

  /* get mode */
  inregs.h.ah = 0xf;
  int86(0x10, &inregs, &outregs);

  /* return mode */
  return outregs.h.al;

}

int LoadPalette()
{
  union REGS regs;
  struct SREGS segregs;

  regs.h.ah = 0x10;                    /* fuction 10h */
  regs.h.al = 0x12;
  /* subfunction 12h - set block of color registers */
  regs.x.bx = 0;                       /* start with this reg */
  regs.x.cx = NUM_MCGA_COLORS;         /* do this many */
  regs.x.dx = (unsigned int)rgbinfo;   /* offset to array */
  segregs.es = (unsigned int)((long)rgbinfo >> 16);
  /* segment of array */
  int86x(0x10, &regs, &regs, &segregs);/* dump data to color registers */

  return SUCCESS;

}



/* ---------------------------------------------------------------------- */
/* Write a pixel at x,y using color                                       */
/* ---------------------------------------------------------------------- */
void PutPixel(int x, int y,unsigned int pixelvalue, unsigned char *framebuffer)
{
    if (x<XMIN || x>XMAX || y<YMIN || y>YMAX)
      return;

    framebuffer[(y*RASTERWIDTH)+x]=(unsigned char)pixelvalue;
}


void XPixel(int x, int y, unsigned char *framebuffer)
{

    if (x<XMIN || x>XMAX || y<YMIN || y>YMAX)
      return;

    framebuffer[(y*RASTERWIDTH)+x] = framebuffer[(y*RASTERWIDTH)+x]^0xff;
}


void XBox(int x1, int y1, int x2, int y2)
{
   int x, y;

   for (x = x1; x <= x2; x++)XPixel(x, y1, FRAMEADDR);
   for (y = (y1+1); y < y2; y++) {
     XPixel(x1, y, FRAMEADDR);
     XPixel(x2, y, FRAMEADDR);
   }
   for (x = x1; x <= x2; x++)XPixel(x, y2, FRAMEADDR);
   return;
}


void LineBox(int x1, int y1, int x2, int y2, unsigned int pixelvalue)
{
   int x, y;

   for (x = x1; x <= x2; x++)PutPixel(x, y1, pixelvalue, FRAMEADDR);
   for (y = (y1+1); y < y2; y++) {
     PutPixel(x1, y, pixelvalue, FRAMEADDR);
     PutPixel(x2, y, pixelvalue, FRAMEADDR);
   }
   for (x = x1; x <= x2; x++)PutPixel(x, y2, pixelvalue, FRAMEADDR);
   return;
}


void PCRamFont(unsigned char *str,
               int xorigin, int yorigin, int scale,
               int fontcolor, int outlinecolor)
{
    int scanline,
        yreg=yorigin,
        xreg=xorigin,
        byt,
        character,
        nibble,
        color;
    int x,y;

    /* flags etcetera */

    int target = strlen(str);  /* string length */

    for (scanline=0;scanline<CELL_SIZE;scanline++)/* finish the current scanline*/
    {                                     /*  before advancing to the next*/
      for (byt=0;byt<target;byt++)     /* run the scanline*/
      {
        /* get the bitmap  */
        character = ramfont[(str[byt]&0x7f)*CELL_SIZE+scanline];
        for (nibble=0;nibble<CELL_SIZE;nibble++)
        {
          xreg+=scale;
          /* chew the byte to bits and lite the pixel if it's a swallow  */
          if (str[byt]!=CRETURN && str[byt]!=LFEED) {
            if (character & 0x80>>nibble)
              color = fontcolor;
            else
              color = outlinecolor;
            if (color > -1 ) {
              for (x=0; x!=scale; x++)
                for (y=0;y!=scale;y++)
                  PutPixel(xreg+x,yreg+y,color, FRAMEADDR);
            }
          }
        }
      }
      yreg+=scale;
      xreg=xorigin;
    }

}

/* ---------------------------------------------------------------------- */
/* PCMidFont                                                              */
/* Maps to PCRamFont, uses a centre-justified x-coordinate.               */
/* ---------------------------------------------------------------------- */
void PCMidFont(unsigned char *str, int xmiddle, int yorigin,
               int scale, int fontcolor, int outlinecolor)
{   /* centre justified string */
    PCRamFont(str,(xmiddle-(4*(strlen(str))*scale)),yorigin,scale,
              fontcolor, outlinecolor);
}

/* ---------------------------------------------------------------------- */
/* GetMcgaPalette is called during the loading of the input file.         */
/* ---------------------------------------------------------------------- */
unsigned char GetMcgaPaletteIndex(unsigned int cgacolor)
{

  // I just hardcoded these in here for CGA images.
  // I couldn't see any point in getting too fancy with 4 colors
  // There isn't much hardship in toggling 4 colors anyway
  // My main target for automation was when a 16 color BMP is colored
  // in Windows Paint... then exported in this thing.
  unsigned int uiIndex;

  switch(cgacolor) {
    case  CGA_WHITE:
          uiIndex = LOWHITE; break;
    case  CGA_MAGENTA:
          uiIndex = LORED; break;
    case  CGA_CYAN:
          uiIndex = LODKBLUE; break;
    case  CGA_BLACK:
    default:
          uiIndex = LOBLACK; break;
  }
  return (unsigned char)uiIndex;
}


void SetPalette()
{

  unsigned int idx, x, temp;

	memset((unsigned char *)&rgbinfo[0][0], 0, (NUM_MCGA_COLORS*NUM_RGB_COLORS));

	/* Set the lightest and darkest color in the current palette.  */
	/* use the darkest color for the drawcolor                     */
	/* use the lightest color for the outline color                */

	drawcolor    = 254;
	outlinecolor = 255;

	// using 6 bit color model (VGA standard video uses 6 bits)
	// for the internal program. values are in the range 0-63
	rgbinfo[255][0] = rgbinfo[255][1] =  rgbinfo[255][2] = 63;

	// leave the drawcolor and outline colors alone
	for (idx = 0; idx < NUM_VGA_COLORS; idx++) {
		for (x = 0; x < NUM_RGB_COLORS; x++) {
		   temp = rgbArray[idx][x];
		   rgbinfo[idx][x] = temp >> 2; // downshift to 6 bits of color
		}
	}

}


/* the following attempts to match BMP colors with a common palette */
/* that is to say... a basic default EGA style 16 color palette like */
/* the kind that Windows Paint provides as a lowest common denominator */
/* my rationale here is that if we can't automate the entire remapping */
/* we fail the whole sh*terree because otherwise we could start mashing */
/* two colors or more together and I for one would not care for that... */
int GetVGAIndex(unsigned char red, unsigned char green, unsigned char blue)
{

    int idx;

    // try exact match
    for (idx = 0; idx < NUM_VGA_COLORS; idx++) {
		if (red == rgbBmpArray[idx][0] && green == rgbBmpArray[idx][1] && blue == rgbBmpArray[idx][2])return idx;
	}
	for (idx = 0; idx < NUM_VGA_COLORS; idx++) {
		if (red == rgbXmpArray[idx][0] && green == rgbXmpArray[idx][1] && blue == rgbXmpArray[idx][2])return idx;
	}
	for (idx = 0; idx < NUM_VGA_COLORS; idx++) {
        if (red == rgbVgaArray[idx][0] && green == rgbVgaArray[idx][1] && blue == rgbVgaArray[idx][2])return idx;
	}
	for (idx = 0; idx < NUM_VGA_COLORS; idx++) {
        if (red == rgbPcxArray[idx][0] && green == rgbPcxArray[idx][1] && blue == rgbPcxArray[idx][2])return idx;
	}

	// adjust gun values to match using EGA-like thresholds
	// this corresponds to the old PCX style palette
	if (red < 43) red = 0;         // 0x00
	else if (red < 128) red = 85;  // 0x55
	else if (red < 224) red = 170; // 0xaa
	else red = 255;                // 0xff

	if (green < 43) green = 0;
	else if (green < 128) green = 85;
	else if (green < 213) green = 170;
	else green = 255;

	if (blue < 43) blue = 0;
	else if (blue < 128) blue = 85;
	else if (blue < 213) blue = 170;
	else blue = 255;

    // try again
    for (idx = 0; idx < NUM_VGA_COLORS; idx++) {
		if (red == rgbBmpArray[idx][0] && green == rgbBmpArray[idx][1] && blue == rgbBmpArray[idx][2])return idx;
	}
	for (idx = 0; idx < NUM_VGA_COLORS; idx++) {
		if (red == rgbXmpArray[idx][0] && green == rgbXmpArray[idx][1] && blue == rgbXmpArray[idx][2])return idx;
	}
	for (idx = 0; idx < NUM_VGA_COLORS; idx++) {
        if (red == rgbVgaArray[idx][0] && green == rgbVgaArray[idx][1] && blue == rgbVgaArray[idx][2])return idx;
	}
	for (idx = 0; idx < NUM_VGA_COLORS; idx++) {
        if (red == rgbPcxArray[idx][0] && green == rgbPcxArray[idx][1] && blue == rgbPcxArray[idx][2])return idx;
	}
	// if no match yet let them pick lores color manually
    return INVALID;
}


int PaletteIndex[NUM_VGA_COLORS] = { 0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15};

void TogglePalette(unsigned char ch)
{
	/* after the initial remapping of colors by color order */
	/* it is very likely that not all entries will remap correctly */
	/* so the palette can be toggled to adjust this */
	int idx, jdx;

	ch = toupper(ch);

	switch(ch)
	{
		case 'A':
		case 'B':
		case 'C':
		case 'D':
		case 'E':
		case 'F':
		  idx = ch - 55;
		  break;

		default:

		    if (ch < '0' || ch > '9')return;
		    idx = ch - 48;
	}

	/* update remapping index */
    jdx = PaletteIndex[idx] + 1;
    if (jdx > 15)jdx = 0;
    PaletteIndex[idx] = jdx;
    /* update the working array */
    rgbArray[idx][0] = rgbLoresArray[jdx][0];
    rgbArray[idx][1] = rgbLoresArray[jdx][1];
    rgbArray[idx][2] = rgbLoresArray[jdx][2];
    /* update the visual */
    SetPalette();
    LoadPalette();

}


/* ---------------------------------------------------------------------- */
/* File Related Functions                                                 */
/* Image Loaders, Image Savers, etc.                                      */
/* ---------------------------------------------------------------------- */

/* type conversion functions */
unsigned int byteword(unsigned char a, unsigned char b){
  return b << 8 | a;
}
unsigned char lsb(unsigned int word){
  return word &0xff;
}
unsigned char msb(unsigned int word){
  return word >> 8;
}

int CheckForCGAPCX(unsigned char *name)
{
  FILE *fp;
  /* reads a ZSOFT .PCX header but ignores the color map */
  int i;
  /* we only want CGA COLOR compatible full screens. */

  unsigned char pcxheader[128];
  unsigned int zsoft,version,codetype,pixbits;
  unsigned int xmin, ymin, xmax, ymax;
  unsigned int x, y;
  unsigned int no_planes, bytesperline;
  int status = VALID;

  /* read the file header */
  if((fp = fopen(name, "rb")) == NULL)return INVALID;
  for(i = 0;i < 128;i++)pcxheader[i] = fgetc(fp);
  fclose(fp);

  zsoft = pcxheader[0];
  version = pcxheader[1];
  codetype = pcxheader[2];
  pixbits = pcxheader[3];

  if(zsoft != 10)
    status = INVALID;
  if(codetype != 1)
    status = INVALID;
  if(pixbits != 2)        /* accept only CGA color images */
    status = INVALID;     /* monochrome images can't be mapped properly */

  xmin = byteword(pcxheader[4], pcxheader[5]);
  ymin = byteword(pcxheader[6], pcxheader[7]);
  xmax = byteword(pcxheader[8], pcxheader[9]);
  ymax = byteword(pcxheader[10], pcxheader[11]);

  no_planes = pcxheader[65];
  bytesperline = byteword(pcxheader[66], pcxheader[67]);

  x =  xmax - xmin;
  y =  ymax - ymin;

  if(x != XMAX)status = INVALID;
  if(y != YMAX)status = INVALID;
  if(no_planes != 1)status = INVALID;
  if(bytesperline != 80)status = INVALID;    /* full screens only */

  /* we can ignore the color map since we        */
  /* are limiting ourselves to CGA modes         */
  /* so we will not handle over 2-bits per pixel */

  return status;

}

int BSAVE_Read(unsigned char *name)
{
  int fh,fh2,y,i;
  unsigned char byte;
  unsigned char *crt;
  unsigned char namebuf[128];
  unsigned char headbuf[7];
  unsigned char linebuffer[80];
  long target = 16384l,target2,header = 7;

  /* open the file twice,eliminate the interleaf,read contiguous */
  sprintf(namebuf, "%s.BAS", name);

  if((fh = open(namebuf, O_RDONLY | O_BINARY)) == - 1)return INVALID;

  if((fh2 = open(namebuf, O_RDONLY | O_BINARY)) == - 1) {
    close(fh);
    return INVALID;
  }

  read(fh, headbuf, sizeof(BSAVED_header));

  /* only read the first 3 bytes, some of these are a little */
  /* different size, depending on how they were saved.       */

  for(i = 0;i < 3;i++)
    if(headbuf[i] != BSAVED_header[i]) {
      close(fh);
      close(fh2);
      return INVALID;
    }

  target2 = (target / 2) + header;
  lseek(fh, (long)(header), SEEK_SET)  ;
  lseek(fh2, (long)(target2), SEEK_SET);

  crt = (unsigned char *)&rawbuffer[0];

  /* translate from a 2 bit color pixel to an 8 bit color pixel */

  for(y = 0;y < SCREENHEIGHT;y += 2) {
    read(fh, linebuffer, 80);
    for(i = 0;i < 80;i++) {
      byte = linebuffer[i];
      *crt++ = GetMcgaPaletteIndex((byte >> 6));
      *crt++ = GetMcgaPaletteIndex((byte >> 4)&3);
      *crt++ = GetMcgaPaletteIndex((byte >> 2)&3);
      *crt++ = GetMcgaPaletteIndex((byte)&3);
    }
    read(fh2, linebuffer, 80);
    for(i = 0;i < 80;i++) {
      byte = linebuffer[i];
      *crt++ = GetMcgaPaletteIndex((byte >> 6));
      *crt++ = GetMcgaPaletteIndex((byte >> 4)&3);
      *crt++ = GetMcgaPaletteIndex((byte >> 2)&3);
      *crt++ = GetMcgaPaletteIndex((byte)&3);
    }
  }
  close(fh);
  close(fh2);
  return SUCCESS;

}

/* Translation for VGA to display CGA */
int PCX_Read(unsigned char *pcxfilename)
{
  unsigned char pcxheader[128];
  unsigned char *crt;
  unsigned int packet;
  unsigned char bit[4];
  FILE *fp;
  unsigned char byte,bytecount;
  long wordcount,target;
  unsigned char name1[128], name2[128];

  sprintf(name1, "%s.PCX", pcxfilename);

  if(CheckForCGAPCX(name1) == - 1)return INVALID;

  if (NULL == (fp = fopen(name1, "rb")))return INVALID;

  target = filelength(fileno(fp));
  for(wordcount = 0;wordcount != 128;wordcount++)byte = fgetc(fp);

  crt = (unsigned char *)&rawbuffer[0];

  do {
    bytecount = 1;                     /* start with a seed count */
    byte = fgetc(fp);
    wordcount++;
    /* check to see if its raw */
    if(0xC0 == (0xC0 &byte)) {
      /* if its not, run encoded */
      bytecount = 0x3f &byte;
      byte = fgetc(fp);
      wordcount++;
    }
    /* translate from 2 bit pixel to 8 bit pixel */

    bit[0] = GetMcgaPaletteIndex((byte >> 6));
    bit[1] = GetMcgaPaletteIndex((byte >> 4)&3);
    bit[2] = GetMcgaPaletteIndex((byte >> 2)&3);
    bit[3] = GetMcgaPaletteIndex((byte)&3);
    packet = 0;
    while(packet++ < bytecount) {
      *crt++ = bit[0];
      *crt++ = bit[1];
      *crt++ = bit[2];
      *crt++ = bit[3];
    }
  }while(wordcount < target);
  fclose(fp);
  return(0);
}

unsigned char bmpscanline[480];

int BMP16_Read(unsigned char *basename)
{

    int width, height, packet = INVALID;
	unsigned int temp, temp2;
	unsigned char *screenbuffer, bmpfile[128];
	int x, x1, y;
	FILE *fp;

    sprintf(bmpfile,"%s.BMP",basename);
	fp=fopen(bmpfile,"rb");

	if (NULL == fp) return INVALID;

    memset((char *)&BitMapFileHeader.bfType, 0, sizeof(BITMAPFILEHEADER));
    memset((char *)&bmi.biSize, 0, sizeof(BITMAPINFOHEADER));

    /* read the header stuff into the appropriate structures */
    fread((char *)&BitMapFileHeader.bfType,
	             sizeof(BITMAPFILEHEADER),1,fp);
    fread((char *)&bmi.biSize,
                 sizeof(BITMAPINFOHEADER),1,fp);
    /* restricted to a reasonable number of formats
       that make sense to me for such a small resolution
       and essentially "cartoon colors" */
    if (bmi.biCompression==BI_RGB &&
        BitMapFileHeader.bfType[0] == 'B' && BitMapFileHeader.bfType[1] == 'M' &&
        bmi.biPlanes==1 && bmi.biBitCount == 4) {
		width = (int) bmi.biWidth;
		height = (int) bmi.biHeight;

        if (width < 321 && height < 201) {
		    packet = width/2;
		    if ((width%2) != 0) packet++;
		    while ((packet%4) != 0)packet++;
		}
	}

    if (packet == INVALID) {
		fclose(fp);
		return INVALID;
	}

   	for(y=0;y<NUM_VGA_COLORS;y++)
	{
	  /* RGB Quad Structure b,g,r,0 */
	  for(x=3;x>0;x--)
	  {
		temp = fgetc(fp);
		rgbOriginalArray[y][x-1] = temp;

	  }
	  fgetc(fp); /* toss rgbReserved */
	}

    for(y=0;y<NUM_VGA_COLORS;y++) {
		// attempt to reorganize remapping of the BMP
		// to the apple lores palette based on rgb values
		x = GetVgaIndex(rgbOriginalArray[y][0],
						rgbOriginalArray[y][1],
						rgbOriginalArray[y][2]);
		// however, if all colors do not match then
		// we simply use the default color order
		// and let them manually select the apple lores color
		if (x == INVALID) {
		    for (x = 0; x < NUM_VGA_COLORS; x++)
		       VGAtoApple[x] = VGAtoAppleSaved[x]; 	// reset
		    break;

		}
		/* move x to y for the remap of the image data */
		/* if they don't like it they can toggle the colors */
	    VGAtoApple[y] = VGAtoAppleSaved[x];

	}

    /* remap the image to the equivalent apple2 lores colors */
    /* to the best known equivalents */
	for (y = height; y > 0; y--) {
	  fread((char *)&bmpscanline[0],1,packet,fp);
	  screenbuffer=(unsigned char *)&rawbuffer[((y-1)*SCREENWIDTH)];
	  for (x = 0, x1=0; x < width; x++) {
		if (x % 2 == 0) {
		  temp  = bmpscanline[x1]; x1++;
		  temp2 = (temp >> 4);
		}
		else {
		  temp2 = (temp & 15);
		}
        screenbuffer[x] = VGAtoApple[temp2];

	  }
	}

	fclose(fp);

  	return SUCCESS;

}

/* routines to save to Apple 2 Lores Format */

/* base addresses for Apple II primary text page */
/* also the base addresses for the 48 scanline pairs */
/* for Apple II lores graphics mode 40 x 48 x 16 colors */
int textbase[24]={
    0x0400,
    0x0480,
    0x0500,
    0x0580,
    0x0600,
    0x0680,
    0x0700,
    0x0780,
    0x0428,
    0x04A8,
    0x0528,
    0x05A8,
    0x0628,
    0x06A8,
    0x0728,
    0x07A8,
    0x0450,
    0x04D0,
    0x0550,
    0x05D0,
    0x0650,
    0x06D0,
    0x0750,
    0x07D0};

#define RAGWIDTH  80
#define RAGHEIGHT 48
#define RAGSIZE   1920
#define TOPSIZE   1600
#define BINSIZE   1016
unsigned char lobuf[RAGSIZE];

/* sets the pixels in the lores buffer (lobuf) */
void setlopixel(unsigned char color,int x, int y,int ragflag)
{
     unsigned char *crt, c1, c2;
     int y1, offset;

     y1 = y / 2;

     c2 = (unsigned char ) (color & 15);

     if (y%2 == 0) {
		 /* even rows in low nibble */
		 /* mask value to preserve high nibble */
		 c1 = 240;
	 }
	 else {
		 /* odd rows in high nibble */
		 /* mask value to preserve low nibble */
		 c1 = 15;
		 c2 = c2 * 16;
	 }

     if (ragflag)
		 offset = (y1 * 80) + x;
     else
		 offset = (textbase[y1]-1024)+x;

	 crt = (unsigned char *)&lobuf[offset];
     crt[0] &= c1;
     crt[0] |= c2;
}


/* save 2 output files */
int savelofragment(unsigned char *basename, int x1, int y1)
{

	FILE *fp;
	unsigned char outfile[128], temp, remap;
	int x,y,x2,y2, offset;

   /* save single lo-res and double lo-res */
    /* save raster images of 48 or 40 scanlines
       (full graphics or mixed text and graphics) */

    if (lores == TRUE) {
		if (appletop == TRUE)
			sprintf(outfile,"%s.STO",basename);
		else
			sprintf(outfile,"%s.SLO",basename);
	}
	else {
		if (appletop == TRUE)
			sprintf(outfile,"%s.DTO",basename);
		else
			sprintf(outfile,"%s.DLO",basename);
	}

    fp = fopen(outfile,"wb");
	if (NULL == fp)return INVALID;

    // On the double lo res display each byte in
    // high memory is interleaved with a byte in low memory
    // in the interests of efficiency I am saving and loading
    // the interleaf on a scanline by scanline basis.
	memset(lobuf,0,RAGSIZE);
	for (y = 0; y< 48; y++) {
		if (appletop == TRUE && y > 39)break;
		y2 = y + y1;
		// first 40 bytes goes to auxilliary memory (even pixels)
		for (x = 0; x < 40; x++) {
			x2 = (x*2) + x1;
			temp = getpixel(x2,y2);
			remap =  PaletteIndex[temp];
			temp = dloauxcolor[remap];
			setlopixel(temp,x,y,1);
		}
		// followed by the interleaf (odd pixels)
		// next 40 bytes goes to main memory
		for (x = 0; x < 40; x++) {
			x2 = (x*2) + x1 + 1;
			temp = getpixel(x2,y2);
			remap =  PaletteIndex[temp];
			setlopixel(remap,x+40,y,1);
		}
	}
	if (lores == TRUE) {
	    fputc(40,fp); // bytes
	    if (appletop == TRUE) fputc(20,fp);
	    else fputc(24,fp);

	    for (y = 0; y < 24; y++) {
			if (appletop == TRUE && y > 19)break;
 	 		offset = (y * 80)+40;
	 		fwrite((unsigned char *)&lobuf[offset],1,40,fp);
		}
	}
	else {
		fputc(80,fp); // bytes
		if (appletop == TRUE) {
			fputc(20,fp); // bytes (rasters / 2)
			fwrite(lobuf,1,TOPSIZE,fp);
		}
		else {
			fputc(24,fp); // bytes (rasters / 2)
			fwrite(lobuf,1,RAGSIZE,fp);
		}
	}
	fclose(fp);

    // for double lo-res */
    // the bsaved images are split into two files
    // the first file is loaded into aux mem
    if (lores == FALSE) {
		sprintf(outfile,"%s.DL1",basename);
		fp = fopen(outfile,"wb");
		if (NULL == fp)return INVALID;


		memset(lobuf,0,BINSIZE);
		for (y = 0; y< 48; y++) {
			y2 = y + y1;
			for (x = 0; x < 40; x++) {
				x2 = (x*2) + x1;
				temp = getpixel(x2,y2);
				remap =  PaletteIndex[temp];
				temp = dloauxcolor[remap];
				setlopixel(temp,x,y,0);
			}
		}
		fwrite(lobuf,1,BINSIZE,fp);
		fclose(fp);
	}

    // for single lo res only 1 file is needed
    // for double lo res
    // the second file is loaded into main mem
    if (lores == TRUE)
    	sprintf(outfile,"%s.SL2",basename);
    else
    	sprintf(outfile,"%s.DL2",basename);

	fp = fopen(outfile,"wb");
	if (NULL == fp)return INVALID;

	memset(lobuf,0,BINSIZE);
	for (y = 0; y< 48; y++) {
		y2 = y + y1;
		for (x = 0; x < 40; x++) {
			x2 = (x*2) + x1 + 1;
			temp = getpixel(x2,y2);
			remap =  PaletteIndex[temp];
			setlopixel(remap,x,y,0);
		}
	}
	fwrite(lobuf,1,BINSIZE,fp);
	fclose(fp);

	return SUCCESS;
}


/* ------------------------------------------------------------------------ */
/* User Input and helper functions and Main Program                         */
/* ------------------------------------------------------------------------ */

// eat keystrokes... avoid mindlessly cycling through the program
// if the user leans on the keyboard and doesn't budge for a moment.
// Any DOS program should do this...

int EatKeys()
{
  if (kbhit())
    while (kbhit())
      if (getch() == FUNCKEY)
         getch();

  return SUCCESS;
}

// show the title at startup and when 'H' (Help) is pressed.
void ShowTitle()
{
    int y, yorg, y1, x1, xx, c;
    unsigned char *ptr;

    for (y = 0; szTitle[y]!= NULL; y++);
    y+=2;


    yorg = y1 = ((SCREENHEIGHT - y*CELL_SIZE) / 2) - 1;
    for (y = 0; szTitle[y]!= NULL; y++) {
      ptr = (char *)&szTitle[y][0];
      PCMidFont(ptr, XMOS, y1, 1, drawcolor, outlinecolor);
      y1+=CELL_SIZE;
    }
    PCMidFont(ptr, XMOS, y1, 1, drawcolor, outlinecolor);
    PCMidFont(ptr, XMOS, y1+CELL_SIZE, 1, drawcolor, outlinecolor);

    // a little finishing touch for the help screen
    // to show the currently selected colors.
    // the following would have been just as tedious no matter
    // how it was programmed... but it looks OK.
    xx = XMOS-136;
    c = 1;
	PCMidFont("C", (xx+=8), y1, 1, (c++), drawcolor);
	PCMidFont("u", (xx+=8), y1, 1, (c++), drawcolor);
	PCMidFont("r", (xx+=8), y1, 1, (c++), drawcolor);
	PCMidFont("r", (xx+=8), y1, 1, (c++), drawcolor);
	PCMidFont("e", (xx+=8), y1, 1, (c++), drawcolor);
	PCMidFont("n", (xx+=8), y1, 1, (c++), drawcolor);
	PCMidFont("t", (xx+=8), y1, 1, (c++), drawcolor);
	PCMidFont(" ", (xx+=8), y1, 1, (c++), drawcolor);
	PCMidFont("C", (xx+=8), y1, 1, (c++), drawcolor);
	PCMidFont("o", (xx+=8), y1, 1, (c++), drawcolor);
	PCMidFont("l", (xx+=8), y1, 1, (c++), drawcolor);
	PCMidFont("o", (xx+=8), y1, 1, (c++), drawcolor);
	PCMidFont("r", (xx+=8), y1, 1, (c++), drawcolor);
	PCMidFont("s", (xx+=8), y1, 1, (c++), drawcolor);
	PCMidFont(":", (xx+=8), y1, 1,  c,    drawcolor);
	PCMidFont(" ", (xx+=8), y1, 1,  c,    drawcolor);
    c = 0;
    PCMidFont("0", (xx+=8), y1, 1, (c++), outlinecolor);
    PCMidFont("1", (xx+=8), y1, 1, (c++), drawcolor);
    PCMidFont("2", (xx+=8), y1, 1, (c++), drawcolor);
    PCMidFont("3", (xx+=8), y1, 1, (c++), drawcolor);
    PCMidFont("4", (xx+=8), y1, 1, (c++), drawcolor);
    PCMidFont("5", (xx+=8), y1, 1, (c++), drawcolor);
    PCMidFont("6", (xx+=8), y1, 1, (c++), drawcolor);
    PCMidFont("7", (xx+=8), y1, 1, (c++), drawcolor);
    PCMidFont("8", (xx+=8), y1, 1, (c++), drawcolor);
    PCMidFont("9", (xx+=8), y1, 1, (c++), drawcolor);
    PCMidFont("A", (xx+=8), y1, 1, (c++), drawcolor);
    PCMidFont("B", (xx+=8), y1, 1, (c++), drawcolor);
    PCMidFont("C", (xx+=8), y1, 1, (c++), drawcolor);
    PCMidFont("D", (xx+=8), y1, 1, (c++), drawcolor);
    PCMidFont("E", (xx+=8), y1, 1, (c++), drawcolor);
    PCMidFont("F",  xx,     y1, 1,  c,    drawcolor);
    y1+=CELL_SIZE;
    y1+=CELL_SIZE;

    x1 = strlen(ptr) * 4;
    LineBox(XMOS - x1 + 2, yorg+1, XMOS + x1 - 1, y1-2, drawcolor);

    EatKeys();
    while (!kbhit());
    EatKeys();

}


/* 24 bit bmp routines - these are used for a non-interactive blind-save
   the 24 bit BMP's that are processed by the following routines are a
   "size-on" double-scaled Double Lo-Res Screen with a mapped palette
   (hopefully) so no need to clip a part of them so no need for our
   clipping editor.

*/
unsigned char GetDLGRColor(unsigned char red, unsigned char green, unsigned char blue)
{
	int idx;

	/* try everything within reason to map to double lo-res colors
	   without guessing *

    /* default to 1.22 colors for now */
	for (idx = 0; idx < 16; idx++) {
		if (red == awinnewcolors[idx][0] && green == awinnewcolors[idx][1] && blue == awinnewcolors[idx][2])
		  return awinnewcolors[idx][3];
	}

	/* try 1.16 colors next */
	for (idx = 0; idx < 16; idx++) {
		if (red == awinoldcolors[idx][0] && green == awinoldcolors[idx][1] && blue == awinoldcolors[idx][2])
		  return awinoldcolors[idx][3];
	}

	/* try BMP2LO colors next */
	for (idx = 0; idx < 16; idx++) {
		if (red == rgbLoresArray[idx][0] && green == rgbLoresArray[idx][1] && blue == rgbLoresArray[idx][2])
		  return (unsigned char)idx;
	}

	/* finally try for an EGA style palette match...
	   otherwise it's some other wierd color so
	   paint it black. */

	idx = GetVGAIndex(red,green,blue);
	if (idx > 0) return VGAtoApple[idx];
	return 0;
}


/* an 8 bit 80 x 48 buffer for Lo-Res Color Indices
   (pixels) - this is used for half-scaling 24 bit bmps
   see below... */
unsigned char bmpbuf[3840];

/* gets the pixel in the bmpbuf */
unsigned char getpixel24(int x, int y)
{
	unsigned char pixel;
	int offset;

	if (x < 0 || x > 79) x = 0;
	if (y < 0 || y > 47) y = 0;

	offset = (y * 80) + x;
	pixel = bmpbuf[offset];

	return pixel;

}

/* sets the pixel in the bmpbuf */
void setpixel24(int x, int y, unsigned char pixel)
{
	int offset;

	if (x < 0 || x > 79) x = 0;
	if (y < 0 || y > 47) y = 0;
	if (pixel > 15) pixel  = 0;

	offset = (y * 80) + x;
	bmpbuf[offset] = pixel;
}



/* blind-save 2 types of Apple II output files */
/* I decided to duplicate much of this code in order to
   keep it separate from the save routines used by the
   clipping editor... setting conditionals to
   read and remap from clips of screen memory vs our tiny buffer
   aren't worth the hassle or the confusion to do
   it all in the same function */
int savelofragment24(unsigned char *basename)
{

	FILE *fp;
	unsigned char outfile[128], temp, remap;
	int x,y,x2,offset;

    /* save single lo-res and double lo-res */
    /* save raster images of 48 or 40 scanlines
       (full graphics or mixed text and graphics) */

    if (lores == TRUE) {
		if (appletop == TRUE)
			sprintf(outfile,"%s.STO",basename);
		else
			sprintf(outfile,"%s.SLO",basename);
	}
	else {
		if (appletop == TRUE)
			sprintf(outfile,"%s.DTO",basename);
		else
			sprintf(outfile,"%s.DLO",basename);
	}

	fp = fopen(outfile,"wb");
	if (NULL == fp)return INVALID;

    // On the double lo res display each byte in
    // high memory is interleaved with a byte in low memory
   	memset(lobuf,0,RAGSIZE);
	for (y = 0; y< 48; y++) {
		if (appletop == TRUE && y > 39)break;
		// first 40 bytes goes to auxiliary memory (even pixels)
		for (x = 0; x < 40; x++) {
			x2 = (x*2);
			temp = getpixel24(x2,y);
			remap = dloauxcolor[temp];
			setlopixel(remap,x,y,1);
		}
		// followed by the interleaf (odd pixels)
		// next 40 bytes goes to main memory
		for (x = 0; x < 40; x++) {
			x2 = (x*2) + 1;
			temp = getpixel24(x2,y);
			setlopixel(temp,x+40,y,1);
		}
	}
	if (lores == TRUE) {
	    fputc(40,fp); // bytes
	    if (appletop == TRUE) fputc(20,fp);
	    else fputc(24,fp);

	    for (y = 0; y < 24; y++) {
			if (appletop == TRUE && y > 19)break;
 	 		offset = (y * 80)+40;
	 		fwrite((unsigned char *)&lobuf[offset],1,40,fp);
		}
	}
	else {
		fputc(80,fp); // bytes
		if (appletop == TRUE) {
			fputc(20,fp); // bytes (rasters / 2)
			fwrite(lobuf,1,TOPSIZE,fp);
		}
		else {
			fputc(24,fp); // bytes (rasters / 2)
			fwrite(lobuf,1,RAGSIZE,fp);
		}
	}
	fclose(fp);

    // for double lo-res */
    // the bsaved images are split into two files
    // the first file is loaded into aux mem

    if (lores == FALSE) {
		sprintf(outfile,"%s.DL1",basename);
		fp = fopen(outfile,"wb");
		if (NULL == fp)return INVALID;

		memset(lobuf,0,BINSIZE);
		for (y = 0; y< 48; y++) {
			for (x = 0; x < 40; x++) {
				x2 = (x*2);
				temp = getpixel24(x2,y);
				remap = dloauxcolor[temp];
				setlopixel(remap,x,y,0);
			}
		}
		fwrite(lobuf,1,BINSIZE,fp);
		fclose(fp);
	}

    // for single lo res only 1 file is needed
    // for double lo res
    // the second file is loaded into main mem
    if (lores == TRUE)
    	sprintf(outfile,"%s.SL2",basename);
    else
    	sprintf(outfile,"%s.DL2",basename);

	fp = fopen(outfile,"wb");
	if (NULL == fp)return INVALID;

	memset(lobuf,0,BINSIZE);
	for (y = 0; y< 48; y++) {
		for (x = 0; x < 40; x++) {
			x2 = (x*2) + 1;
			temp = getpixel24(x2,y);
			setlopixel(temp,x,y,0);
		}
	}
	fwrite(lobuf,1,BINSIZE,fp);
	fclose(fp);

	return SUCCESS;
}


/* any 24 bit bmp in our supported sizes can enter here
   but unless we can map the colors to ours unmapped
   colors will be black in the Apple II Output

   I have protected this routine somewhat by using
   a nominal size of 160 x 96 for image data so unless
   a user is really goofy or didn't bother to
   read the code or the documentation, the results
   should be accurate.

   */
int BMP24_Read(unsigned char *basename)
{

 	FILE *fp;
    int width, height, xoff=0, yoff=0, x, x1, y, i, packet=INVALID;
    unsigned char bmpfile[128],tempr, tempg, tempb, pixel, skip=0;

	sprintf(bmpfile,"%s.BMP",basename);
    if((fp=fopen(bmpfile,"rb"))==NULL) {
		return INVALID;
	}

    memset((char *)&BitMapFileHeader.bfType, 0, sizeof(BITMAPFILEHEADER));
    memset((char *)&bmi.biSize, 0, sizeof(BITMAPINFOHEADER));

    /* read the header stuff into the appropriate structures */
    fread((char *)&BitMapFileHeader.bfType,
	             sizeof(BITMAPFILEHEADER),1,fp);
    fread((char *)&bmi.biSize,
                 sizeof(BITMAPINFOHEADER),1,fp);

    /* restricted to a reasonable number of formats
       that make sense to me for such a small resolution
       and essentially "cartoon colors" */
    if (bmi.biCompression==BI_RGB &&
        BitMapFileHeader.bfType[0] == 'B' && BitMapFileHeader.bfType[1] == 'M' &&
        bmi.biPlanes==1 && bmi.biBitCount == 24) {
		width = (int) bmi.biWidth;
		height = (int) bmi.biHeight;

        /* double scaled and double scaled with "palette" at bottom */
		if (width == 160 && (height == 96 || height == 128)) {
			packet = width*3; /* rgb triples */
			if (height == 128) yoff = 32;
			skip = 1;
		}

		/* "size-on" bmps - like big icons */
		if (width == 80 && (height == 48 || height == 64)) {
		    packet = width*3; /* rgb triples */
		    if (height == 64) yoff = 16;
		}

        /* old and new printshop "pastes" from ClipShop in
           small-copy (single-scaled) or regular copy (double-scaled) format...
           output will be centered... if they don't like
           that they can clip them to "size-on" while they're coloring in Paint
           before converting... hopefully they'll at least know how to do that. */
		if (width == 88 && height == 52) {
		    packet = width*3; /* rgb triples */
		    xoff = 12;
		    yoff = 2;
		}
		if (width == 176 && height == 104) {
		    packet = width*3; /* rgb triples */
		    xoff = 24;
		    yoff = 4;
		    skip = 1;
		}
		/* this is enough for now */

	}

    if(packet == INVALID){
		fclose(fp);
		return INVALID;
	}

 	memset(&bmpbuf[0],0,3840); /* clear write buffer */

    for (i = 0; i < yoff; i++) fread((char *)&bmpscanline[0],1,packet,fp);

    /* map rgb triples to lo-res colors */
    for(i=48,y=47;i>0;i--,y--)
    {
	  /* if skipping, process every 2nd line */
	  if (skip != 0)fread((char *)&bmpscanline[0],1,packet,fp);
	  fread((char *)&bmpscanline[0],1,packet,fp);

	  for (x = 0, x1=xoff; x < 80; x++) {
		 /* read the first triple */
	     tempb = bmpscanline[x1]; x1++;
	     tempg = bmpscanline[x1]; x1++;
	     tempr = bmpscanline[x1]; x1++;
	      /* skip the next triple */
	     if (skip != 0) x1+= 3;

         /* get the lo-res color */
         pixel = GetDLGRColor(tempr,tempg,tempb);
         /* set the pixel in the "compressed" buffer to
            a lo-res color index */
	     setpixel24(x,y,pixel);

	  }
    }
    fclose(fp);

return SUCCESS;
}


void main(int argc, char **argv)
{

  char *tf[] = {"FALSE","TRUE"};
  int status = 0;

  /* bounds of lores image */
  int oldx1=0, oldy1=0, oldx2=79, oldy2=47;
  int x1=0, y1=0, x2=79, y2=47, len, len1;

  unsigned char c, kdx;
  unsigned char fname[128],sname[128],outfile[128];
  unsigned char *wordptr;
  unsigned char buf[128];
  FILE *fp;

  system("cls");
  puts(szTextTitle);

  if(argc == 1) {
	puts("Non-Interactive Conversion    24Bit.BMP (sizes below - supported colors only)");
	puts(" (\"screens\" 160 x 96, 160 x 128, 80 x 48 and \"minipix\" 88 x 52, 176 x 104)");
	puts("Interactive Conversion        CGA.PCX, CGA.BAS, 16Color.BMP");
	puts(" (\"minipix\" clips - \"9 clip canvas\" or 16 color \"fragment\" - sizes below)");
    puts("Command line Usage is \"BMP2LO MyCGA.PCX\"     (320 x 200 9 clip canvas)");
    puts("       80 x 48 output \"BMP2LO MyBSAVE.BAS\"   (320 x 200 9 clip canvas)");
    puts("                      \"BMP2LO My16Color.BMP\" (320 x 200 max fragment size)");
    puts(" MS-DOS 8.3 style     \"BMP2LO My24Bit.BMP\"   (assorted sizes listed at top)");
    puts(" File Names Only      \"BMP2LO MyCGA.PCX OutfileBaseName\"");
    puts("                      \"BMP2LO MyBSAVE.BAS OutfileBaseName\"");
    puts("                      \"BMP2LO My16Color.BMP OutfileBaseName\"");
    puts("                      \"BMP2LO My24Bit.BMP OutfileBaseName\"");
    puts("Option:80 x 40 output \"BMP2LO MyInputfile T\" (Top - mixed mode image)");
    puts("                      \"BMP2LO MyInputfile OutfileBaseName T\"");
    puts("Option:40 x 48 output \"BMP2LO My24Bit.BMP L\" (Low - single lo-res)");
    puts(" (24 bit BMP's only)  \"BMP2LO My24Bit.BMP OutfileBaseName L\"");
    puts("Both Options          \"BMP2LO My24Bit.BMP L T\"");
    puts(" (24 bit BMP's only)  \"BMP2LO My24Bit.BMP OutfileBaseName L T\"");
    printf("Enter Input FileName (Blank to Exit): ");
    gets(fname);
    if (fname[0] == ASCIIZ)
      exit(1);
    printf("Enter OutputFile BaseName (Blank for Default) : ");
    gets(outfile);
  }
  else {
    for (;;) {
		outfile[0] = ASCIIZ;
		strcpy(fname, argv[1]);
	    if (argc == 2) break;

      	strcpy(buf, argv[2]);
      	len = strlen(buf);
      	c = toupper(buf[0]);
      	if (len == 1 && c == 'T')appletop = TRUE;
      	else if (len == 1 && c == 'L')lores = TRUE;
      	else strcpy(outfile,buf);
      	if (argc == 3) break;

  	    strcpy(buf, argv[3]);
      	len = strlen(buf);
      	c = toupper(buf[0]);
      	if (len == 1 && c == 'T' && appletop == FALSE)appletop = TRUE;
      	else if (len == 1 && c == 'L' && lores == FALSE)lores = TRUE;
      	else if (outfile[0] == ASCIIZ)strcpy(outfile,buf);
      	if (argc == 4) break;

		strcpy(buf, argv[4]);
		len = strlen(buf);
		c = toupper(buf[0]);
		if (len == 1 && c == 'T' && appletop == FALSE)appletop = TRUE;
		else if (len == 1 && c == 'L' && lores == FALSE)lores = TRUE;
		else if (outfile[0] == ASCIIZ)strcpy(outfile,buf);
		if (argc == 5) break;

	    break;
	}

	if (appletop == TRUE || lores == TRUE)
	   printf("Current Options: T=%s (40 line output) * L=%s (lo-res output)\n",tf[appletop],tf[lores]);
  }

  if((rawbuffer = malloc((unsigned)65000)) == NULL) {
    puts("Out of Memory...");
    exit(1);
  }
  memset(&rawbuffer[0],0,(unsigned)65000);

  strcpy(sname, fname);
  wordptr = strtok(sname, ".");
  if (outfile[0] == ASCIIZ)strcpy(outfile,sname);

  status = PCX_Read(sname);
  if(status)status = BSAVE_Read(sname);
  if(status)status = BMP16_Read(sname);
  if (status) {

	  status = BMP24_Read(sname);
	  if (status) {
		printf("%s is an Unsupported Format or cannot be opened.\n", fname);
		free(rawbuffer);
		exit(1);
	  }

	  status = savelofragment24(outfile);

	  if (lores == TRUE) {
		  if (appletop == TRUE) sprintf(buf,"%s.SL2 and %s.STO",outfile,outfile);
		  else sprintf(buf,"%s.SL2 and %s.SLO",outfile,outfile);
	  }
	  else {
		  if (appletop == TRUE) sprintf(buf,"%s.DL1 & .DL2 and %s.DTO",outfile,outfile);
		  else sprintf(buf,"%s.DL1 & .DL2 and %s.DLO",outfile,outfile);
	  }

      if (status == SUCCESS)printf("%s Saved!\n",buf);
      else printf("Error saving %s.\n",buf);
  	  free(rawbuffer);
  	  puts("Have a Nice Dos!");
      exit(0);
  }


  /* if we are not processing "blind output" start the clipping editor */
  /* for systems that do not support MS-DOS graphics modes,
     blind output from 24 bit BMPs in our supported size-range and
     colors is the only option */
  status = ESCKEY;

  if((SetCrtMode(MCGA)) == MCGA) {
    SetPalette();
    LoadPalette();
    vload();
    ShowTitle();
    vload();
    xbox(oldx1, oldy1, oldx2, oldy2);
    do {
      c=toupper(getch());

      if (FUNCKEY == c) {
        kdx=getch();
        switch(kdx) {
		  case HOMEKEY:
		       x1 = 0;
		  case LTARROW:
		       x1-=1;
		       x2-=1;
		       if (x1 < 0) {
				 x1 = 0;
				 x2 = 79;
		       }
		       break;
          case ENDKEY:
               x2=319;
          case RTARROW:
		       x1+=4;
		       x2+=4;
		       if (x2 > 319) {
				 x2 = 319;
				 x1 = 319 - 79;
		       }
		       break;
          case PGUP:
               y1 = 0;
          case UPARROW:
		       y1-=1;
		       y2-=1;
		       if (y1 < 0) {
				 y1 = 0;
				 y2 = 47;
		       }
		       break;
          case PGDOWN:
               y2 = 199;
          case DOWNARROW:
		       y1+=4;
		       y2+=4;
               if (y2 > 199) {
			      y2 = 199;
			      y1 = 199 - 47;
			   }

		       break;
          default:
            break;
        }
      }
      else {

        if (c == 'S' || c == ENTERKEY)
          break;

        switch (c) {
          case 'H':
            xbox(oldx1, oldy1, oldx2, oldy2);
            ShowTitle();
            vload();
            xbox(oldx1, oldy1, oldx2, oldy2);
            break;
          case 'Q':
            c = ESCKEY;
            break;
          default:
            if (c < '0' || c > 'F') break;
            if (c > '9' && c < 'A') break;
            TogglePalette(c);
        }
      }

      if (x1!=oldx1 || y1!=oldy1) {
	     /* erase old box */
	     xbox(oldx1, oldy1, oldx2, oldy2);
	     /* update old cords with new cords */
	     oldx1 = x1; oldy1 = y1; oldx2 = x2; oldy2 = y2;
	     /* plot new box */
	     xbox(oldx1, oldy1, oldx2, oldy2);
	  }


    } while (c!=ESCKEY);

    if(c!=ESCKEY)
      status = savelofragment(outfile,x1,y1);
    else
      status = ESCKEY;

    SetCrtMode(TEXT);

  }

  puts(szTextTitle);

  if (status != ESCKEY) {

	  if (appletop == TRUE) sprintf(buf,"%s.DL1 & .DL2 and %s.DTO",outfile,outfile);
	  else sprintf(buf,"%s.DL1 & .DL2 and %s.DLO",outfile,outfile);

      if (status == SUCCESS)printf("%s Saved!\n",buf);
      else printf("Error saving %s.\n",buf);

  }
  else printf("Exiting without saving...\n");

  free(rawbuffer);
  puts("Have a Nice Dos!");
  exit(0);

}
