/*  ------------------------------------------------------------------------
System       : Manx Aztec C65 Version 3.2b
               MS-DOS cross-development environment
Platform     : Apple IIe 128K PRODOS 8
Program      : hlode(C) Copyright Bill Buckels 2013.
               All rights reserved.

Description  : Apple //e ProDOS 8 HGR Slideshow Program

			   HGR formats supported:

			   BSAVED HGR Images - .BIN Files
			   HGR Raw Raster Images - .TOP, .BOT, and .RAG Files
			   HGR Run-Length Encoded Images - .RAX Files

               Uses Page 2 Hi-Res to avoid
               ProDOS Sys Program Load Address at 0x2000

Written by   : Bill Buckels
Date Written : February 2013

Revision     : 1.0 First Release
Licence      : You may use this program for whatever you wish as long
               as you agree that Bill Buckels has no warranty or
               liability obligations whatsoever from said use.
------------------------------------------------------------------------ */

#include <g3.h>
#include <stdio.h>
#include <prodir.h>

#include "piclist.inc"

extern unsigned char XPIC;
extern unsigned char YPIC;

unsigned char *raxreadbuf = (unsigned char *)4192;


/*
	CLR80VID = $C00C ;disable 80-column display mode (WR-only)
 	SET80VID = $C00D ;enable 80-column display mode (WR-only)

unsigned char *clr80vid = (unsigned char *)0xc00c;
unsigned char *set80vid = (unsigned char *)0xc00d;

	SET80STORE=$C001 ;80STORE On- enable 80-column memory mapping (WR-only)
	CLR80STORE=$C000 ;80STORE Off- disable 80-column memory mapping (Write)

unsigned char *set80store = (unsigned char *)0xc001;

*/

unsigned char *clr80store = (unsigned char *)0xc000;

#include "imgtype.inc"

main()
{

	FILE *fp;
	int idx,row,status=0,started=0,imagetype,cnt,c,x,y;
	char picname[82];

    hgr();

    while (kbhit())clearkey();
	if((fp=fopen("PIXLIST","r"))==NULL) {
		/* first time through try to create piclist if it does not exist */
		fp = makepiclist();
		if (fp == NULL) status = -1;
	}
    else {
		/* advance on timeout or advance on keypress */
		if (wait(10,XTIME,YTIME)==27) {
			status = -2;
			fclose(fp);
		}
	}
	while (kbhit())clearkey();


	while(status == 0) {
		text();
		crt80();
		scr_clear();
		row = 0;
		scr_curs(row,0);
		puts("hlode (C)Copyright Bill Buckels 2013. All Rights Reserved.");

		status = -1;
		if (started != 0) {
			if((fp=fopen("PIXLIST","r"))==NULL)break;
	    }
	    started = 1;
		/* get the first filename */
		if (fgets(picname,80,fp) == -1) {
			fclose(fp);
			break;
		}
		if (picname[0] < 33 || picname[0] > 122) {
			fclose(fp);
			break;
		}

		cnt = 0;
        /* allow prefix slash and alpha numeric characters as first character on the line */
		while(picname[0] > 32 && picname[0] < 123)
		{

             /*
			   comment lines are preceded by pound signs or
			   exclamation marks... in fact by any character from
			   ascii 33 to ascii 46 in the first column of a piclist
			   line so we skip over those entries.

		       */

            if (picname[0] > 46) {
				idx=0;
				while(picname[idx]!=0)
				{

					/* replace carriage returns, commas and spaces with null
					   commas are valid for load coordinates of image fragments
					   and when they are used, they follow the filename,x,y
					   */
					if(picname[idx] < 46 || picname[idx] > 122){
					   picname[idx]= 0;
					}
					else {
						picname[idx] = toupper(picname[idx]);
						idx++;
					}
				}
				if (picname[0] == 0)break;

		    /* piclist mnemonics of one character followed by an equals sign are
			   supported for slideshow control in hshow so we skip
			   over those entries... */


                if (picname[1] != 61) {
					if (cnt == 0) {
						row+=2;
						scr_curs(row,0);
						puts("Picture Names :");
					}
					scr_curs(row,20);
					puts(picname);
					cnt++;
					row++;
					if (row == 21) break;
				}
		    }
		    picname[0] = 0;
			if (fgets(picname,80,fp) == -1)break;

		}
    	fclose(fp);
    	if (cnt == 0)break;
    	status = 0;

		/* now that we are displaying the picture list we accept input */
		row++;
		scr_curs(row,0);
		puts("Enter Picture Name:");
		scr_curs(row,20);
		gets(picname); /* get the filename */
		strupr(picname);

        /* get the image type and parse the load coordinates if any */
        imagetype = getimagetype(picname,&x,&y);

		/* exit if a blank or meaningless line was entered */
		if (picname[0] == 0)break;

        scr_clear();
		/* turn 80 store off and switch to hgr mode and clear screen */
		*clr80store = 0;
		black();

		switch(imagetype)
		{            /* centre image fragments on the screen
		                if no load coordinates */
			case 4:  c = raxlode(picname,x,y,3584,raxreadbuf);
					 break;
			case 3:  /* centre image fragments on the screen */
					 XPIC=(unsigned char)x;
					 YPIC=(unsigned char)y;
			case 2:
			case 1:  /* do not centre bot and top images */
			         c = piclode(picname);
			         break;
			default: c = bload(picname);
		}

		if (c!=0) {
			text();
			crt80();
		   	scr_clear();
		   	scr_curs(0,0);
		   	if (c == -1) puts("!Can't open pic!");
		   	else puts("!Pic read error!");
		   	scr_curs(1,0);
		   	puts("Press Any Key...");
		   	getch();
	     }
	     else {
			 if(getch()==27)break;
		 }

	}

    text();
	crt80();
    scr_clear();
	scr_curs(0,0);

	if (status == -1) {
		puts("!PIXLIST missing, invalid, or could not be opened or created!");
		scr_curs(1,0);
		puts("Press Any Key...");
		getch();
		scr_clear();
        scr_curs(0,0);
	}
    _exit();
}
