/*  ------------------------------------------------------------------------
System       : Manx Aztec C65 Version 3.2b
               MS-DOS cross-development environment
Platform     : Apple IIe 128K PRODOS 8
Program      : hshow(C) Copyright Bill Buckels 2013.
               All rights reserved.

Description  : Apple //e ProDOS 8 HGR Slideshow Program

			   HGR formats supported:

			   BSAVED HGR Images - .BIN Files
			   HGR Raw Raster Images - .TOP, .BOT, and .RAG Files
			   HGR Run-Length Encoded Images - .RAX Files

               Uses Page 2 Hi-Res to avoid
               ProDOS Sys Program Load Address at 0x2000

Written by   : Bill Buckels
Date Written : February 2013

Revision     : 1.0 First Release
Licence      : You may use this program for whatever you wish as long
               as you agree that Bill Buckels has no warranty or
               liability obligations whatsoever from said use.
------------------------------------------------------------------------ */
#include <g3.h>
#include <stdio.h>
#include <prodir.h>

extern unsigned char XPIC;
extern unsigned char YPIC;

unsigned char *raxreadbuf = (unsigned char *)4192;


/*
	CLR80VID = $C00C ;disable 80-column display mode (WR-only)
 	SET80VID = $C00D ;enable 80-column display mode (WR-only)

unsigned char *clr80vid = (unsigned char *)0xc00c;
unsigned char *set80vid = (unsigned char *)0xc00d;

	SET80STORE=$C001 ;80STORE On- enable 80-column memory mapping (WR-only)
	CLR80STORE=$C000 ;80STORE Off- disable 80-column memory mapping (Write)

unsigned char *set80store = (unsigned char *)0xc001;

*/

unsigned char *clr80store = (unsigned char *)0xc000;

#include "imgtype.inc"

char *title = "hshow";
char *copyright = "(C) Copyright Bill Buckels 2013.All rights reserved.";


int mystuff()
{
	FILE *fp;
	int cnt = 0, idx, valid = 1;
	char *buf = (char *)4192;

	if((fp=fopen("copyread","r"))==NULL) return cnt;


	/* allow 20 lines for copyright and usage */
	for (cnt = 0; cnt < 20;) {
		buf[0] = 0;
		if (fgets(buf,80,fp) == -1) break;

        /* signature line must match */
		if (cnt == 0) {
		    for (idx = 0;copyright[idx] != 0; idx++) {
				if (copyright[idx] != buf[idx]) {
				   valid = 0;
				   break;
				}
			}
			if (valid == 0) break; /* outa' here */
			scr_curs(cnt,0);puts(title);cnt++;
			scr_curs(cnt,0);puts(copyright);cnt++;
		}
		else {
			for (idx = 0;buf[idx] != 0;) {
				if (buf[idx] == 13||buf[idx] == 10||buf[idx]==0) buf[idx] = 0;
				else {
					if (buf[idx] < 32 || buf[idx] > 122) {
						valid = 0;
						break;
					}
					idx++;
				}
			}
		    if (valid == 0) break; /* outa' here */
			scr_curs(cnt,0);puts(buf);cnt++;
		}
	}
	fclose(fp);
	scr_curs(cnt,0);
	return cnt;

}

/* the following is what I eventually ended-up with as a reasonable
   compromise for the structure of a simple slideshow program for
   ProDOS using a picture file list. */

main()
{
	FILE *fp;
	unsigned filesize;
	int idx,status,key=0,c,cnt,imagetype,x,y,delay=10,loops=-1,lcnt=0;
	char picname[82];

	crt80();
	scr_clear();
	scr_curs(0,0);
	*clr80store = 0;
	black();


    /* loop through the slideshow and restart at the beginning until the
       user presses the escape key... */

 	for (;;) {

		status = -1;
		if((fp=fopen("PIXLIST","r"))==NULL) break;
		picname[0] = 0; /* gigo - blank the line before calling fgets */
		/* get the first filename */
		if (fgets(picname,80,fp) == -1) {
			fclose(fp);
			break;
		}
		/* if garbage on the first line outa' here */
		if (picname[0] < 33 || picname[0] > 122) {
			fclose(fp);
			break;
		}

		/* clear keyboard */
		while (kbhit()!=0)clearkey();
		status = cnt = 0;

		/* line validation */
		while(picname[0] > 32 && picname[0] < 123) {

			idx=0;  /* replace carriage return with null, ignore lines beginning
			           with invalid characters etc. */

			while(picname[idx]!=0)
			{
				 /* if I find an ascii value below the comma character value (44)
				    or any above a lower-case z I truncate the line there... */
				 if(picname[idx] < 44 || picname[idx] > 122)picname[idx]= 0;
				 else idx++;
			}

			strupr(picname);

		    if (picname[1] == 61) {
				/* parse slideshow commands

			    d=delaytimeinseconds
			    c=colornumber
			    l=numberofloops before exit */

                imagetype = atoi(&picname[2]);

                if (picname[0] == 'D') {
					if (imagetype > -1) delay = imagetype;
					else delay = 10;
				}
				if (picname[0] == 'L') {
					if (imagetype > -1) loops = imagetype;
					else loops = -1;
				}
				if (picname[0] == 'C') {
					/* clear to color specified in the piclist

						 black    =0
						 green    =1
						 violet   =2
						 white    =3
						 black    =4
						 red      =5
						 blue     =6
						 white    =7
					*/
				   switch(imagetype)
				   {
					   case 1:  green(); break;
					   case 2:  violet();break;
					   case 5:  red();   break;
					   case 6:  blue();  break;
					   case 3:
					   case 7:  white(); break;
					   default: black();
				   }
			   }
			   picname[0] = 0;
			}
			else  {
				/* get the image type and parse the load coordinates if any */
				imagetype = getimagetype(picname,&x,&y);
			}

			if (picname[0] != 0) {

				/* preload keycheck - skip loading the image if
				   the user leans on the keyboard and try to avoid
				   partially loading raster based images because they
				   can look like heck */
				key = kbhit();
				if (key!=0) {
					clearkey();
					cnt++;
				}
				else {


					switch(imagetype)
					{            /* centre image fragments on the screen
									if no load coordinates */
						case 4:  c = raxlode(picname,x,y,3584,raxreadbuf);
								 break;
						case 3:  /* centre image fragments on the screen */
								 XPIC=(unsigned char)x;
								 YPIC=(unsigned char)y;
						case 2:
						case 1:  /* do not centre bot and top images */
								 c = piclode(picname);
								 break;
						default: c = bload(picname);
					}

					if (c > -1) cnt++;

					/* if a load error skip this slide, otherwise... */
					/* advance on timeout or advance or end on keypress */
					if (c==0) key = wait(delay,XTIME,YTIME);

				} /* nobody leaned on the keyboard to advance to the next slide */
			} /* the format for the picture name entry in the piclist is valid */
			else {
				/* picname[0] == 0 - truncated line - began with garbage */
				key = kbhit();
				if (key!=0)clearkey();

			}
			/* if escape key is pressed end slideshow */
			if (key == 27) break;
			picname[0] = 0; /* gigo - make sure no garbage is going in to fgets
			                          trust nothing and no-one */
			if (fgets(picname,80,fp) == -1)break;
		}
		fclose(fp);

		/* if all the pictures are missing or invalid or if
		   the entire piclist is not correctly formatted
		   the count will be zero...
		   */

        lcnt++;
        if (loops != -1) {
		  if (lcnt > loops) key = 27;
		}
		if (cnt == 0)status = -1;
		if (key == 27 || cnt == 0)break;

		/* bottom of loop - start over at the beginning of
		   the slideshow piclist... */

	}

    text();
	crt80();
    scr_clear();
	scr_curs(0,0);

	/* clear keyboard */
	while (kbhit()!=0)clearkey();

	if (status!=0) {
		puts("!PIXLIST missing, invalid, or could not be opened!");
		scr_curs(1,0);
		puts("Press Any Key...");
		getch();
		scr_clear();
		scr_curs(0,0);
  	}

  	/* display the copyright if one exists */
 	if (mystuff() != 0) {
		puts("\nPress Any Key...");
		getch();
		scr_clear();
		scr_curs(0,0);
	}

	_exit();
}





