/* ------------------------------------------------------------------------

hgrServer.c (C) Copyright Bill Buckels 2013.
All Rights Reserved.

Written by   : Bill Buckels
Email        : bbuckels@mts.net

Date Written : March 15, 2013
Date Revised :

Purpose      : Appple II Legacy Graphics PowerToy

Licence Agreement
-----------------

Bill Buckels has no warranty obligations or liability whatsoever resulting from
the distribution and use of hgrServer or its source code and the files it came
with in any way whatsoever.

If you don't agree, remove hgrServer, its source code, and the files it came
with from your computer now.

You may use hgrServer, its source code, and the files it came with for whatever
you wish as long as you agree that Bill Buckels has no warranty or liability
obligations whatsoever from said use.

hgrServer is written in MSC (Microsoft C) and compiled under the Visual Studio
2005 Command Line with the Win32 SDK. If you are reasonably familiar with
Windows Programming in the C Programming Language you shouldn't have much
trouble reading the source code. I have made no attempt at POSIX compliance
whatsoever. This code is standard Windows C which has really not changed
in the last 23 years (since Windows 3.0) and is targeted exclusively at Windows
and AppleWin, and is, like latin, and Aztec C, almost completely non-portable.

------------------------------------------------------------------------ */

// windows header
#include <windows.h>

// microsoft runtime headers
#include <stdlib.h>
#include <stdio.h>
#include <io.h>
#include <string.h>
#include <dos.h>
#include <direct.h>

// our own header
#include "hgrServer.h"

#define XMAX 280
#define YMAX 192

LRESULT CALLBACK WndProc(HWND, UINT, WPARAM, LPARAM);
int ParseCommandLine(LPSTR lpszCmdParam);
int FileExists(char *spec);
int GetLastFileCount(char *spec);
int RenameFiles();

char *lpszAppName = "hgrServer";
char *lpszTitle   = "hgrServer\251 v1.0 Copyright Bill Buckels 2013.";
char MessageText[4096];
char *lpWelcome =
"hgrServer is now loaded and ready to process Apple II BSaved HGR screen memory"
"\nimages (bin files) created (BSaved) by the AppleWin Debugger in the following"
"\nWorking Directory (WD): %s\n\n"
"If this directory is not the AppleWin debugger WD (listed by the debugger"
"\ncommand PWD), close hgrServer and run it again from your debugger WD *or*"
"\nchange the debugger WD to the above directory by reloading your disk image from"
"\nthe directory listed above using the Applewin toolbar.\n\n"
"hgrServer also accepts a WD Name from the commandline..."
"\nUsage is: hgrServer [WD Name].\n\n"
"If you are using debugger scripts (hgr1 and hgr2 or equivalent) to save bin files,"
"\nthey must be in your debugger WD so the debugger can find them and run them"
"\n(i.e. RUN hgr1). hgrServer processes HGR bin files BSaved using the"
"\nAppleWin debugger's default naming only (i.e. BSAVE ,2000,2000). You may run"
"\nmultiple copies of hgrServer. Make sure to close hgrServer when you are done!\n\n"
"You may use hgrServer, its source code,and the files it came with for whatever"
"\nyou wish as long as you agree that Bill Buckels has no warranty or liability"
"\nobligations whatsoever from said use.\n";

HWND hwndParent; // maybe for future use
HINSTANCE hInst;

int xborder=1,yborder=1;

BOOL busyFlag = FALSE;
BOOL quitFlag = FALSE;

// ------------------------------------------------------------------------
// Windows Entry Point (main module)
// ------------------------------------------------------------------------
int APIENTRY WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance,
                    LPSTR lpszCmdParam, int nCmdShow)
     {
     HWND        hwnd ;
     MSG         msg ;
     WNDCLASSEX  wndclass;
     int ycaption, WindowWidth, WindowHeight, xscreen, yscreen, x1,y1;
     // int ymenu;

	 hInst = hInstance; // save instance handle

     ParseCommandLine(lpszCmdParam);

     if (!hPrevInstance)
          {
	      wndclass.cbSize        = sizeof(WNDCLASSEX);
          wndclass.style         = CS_HREDRAW | CS_VREDRAW ; // fixed frame style
          wndclass.lpfnWndProc   = (WNDPROC)WndProc ;
          wndclass.cbClsExtra    = 0 ;
          wndclass.cbWndExtra    = 0 ;
          wndclass.hInstance     = hInstance ;
          wndclass.hIcon         = LoadIcon(hInstance, "MyIcon");
          wndclass.hCursor       = NULL;
          wndclass.hbrBackground = GetStockObject (WHITE_BRUSH) ;
          wndclass.lpszMenuName  = NULL ;
          wndclass.lpszClassName = lpszAppName ;
          wndclass.hIconSm       = LoadImage(hInstance,
		                                     "MyBigIcon",
		                                     IMAGE_ICON,
		                                     16, 16,0);

          if (!RegisterClassEx(&wndclass))
		            {
		                // if RegisterClassEx() is not implemented
		                // try calling RegisterClass().
		                RegisterClass((LPWNDCLASS)&wndclass.style);
          }
	  }


      // xborder is always used as xoffset into window client area - should be 1
      xscreen = GetSystemMetrics(SM_CXFIXEDFRAME);
      WindowWidth=XMAX+xborder+xborder+xscreen+xscreen;

 	  // yborder is always used as yoffset into window client area - should be 1
      yscreen = GetSystemMetrics(SM_CYFIXEDFRAME);
      ycaption = GetSystemMetrics(SM_CYCAPTION);
      // ymenu = GetSystemMetrics(SM_CYMENU);
      // WindowHeight= YMAX+yborder+yborder+yscreen+yscreen+ycaption+ymenu;
      WindowHeight= YMAX+yborder+yborder+yscreen+yscreen+ycaption;

      // put the window onto centre-stage when open
      // they can move it if they want
      x1=(GetSystemMetrics(SM_CXSCREEN)-XMAX)/2 -xborder-xscreen; // top left
      // y1=(GetSystemMetrics(SM_CYSCREEN)-YMAX)/2 -yborder-yscreen-ycaption-ymenu;
      y1=(GetSystemMetrics(SM_CYSCREEN)-YMAX)/2 -yborder-yscreen-ycaption;// top left

    hwndParent = hwnd = CreateWindow (lpszAppName,lpszTitle,
                          WS_MINIMIZE|WS_MINIMIZEBOX|
                          WS_CAPTION|WS_SYSMENU,
                          x1,y1,
                          WindowWidth,WindowHeight,
                          NULL, NULL, hInstance, NULL);


     ShowWindow(hwnd, SW_SHOWMINIMIZED);
     UpdateWindow (hwnd);

     while (GetMessage (&msg, NULL, 0, 0))
     {
          TranslateMessage (&msg) ;
          DispatchMessage (&msg) ;
     }

     hwndParent = NULL;
     return msg.wParam;
}


// ------------------------------------------------------------------------
// Main Message Loop
// ------------------------------------------------------------------------
LRESULT CALLBACK WndProc (HWND hwnd, UINT message,
                          WPARAM wParam, LPARAM lParam){
    HDC hMemoryDC, hdc;
    HBITMAP hBitmap, hOldBitmap;
    BITMAP bitmap;
    PAINTSTRUCT ps;
    WORD wCommand;


    switch (message)
	  {
        case WM_CREATE:

			SetTimer(hwnd,IDT_RENAME,5000, NULL);
			return 0L;

		case WM_TIMER:
            busyFlag = TRUE;
			KillTimer(hwnd,IDT_RENAME);
			if (quitFlag == TRUE) {
				SetTimer(hwnd,IDT_RENAME,10000, NULL);
			}
			else {
			   // I don't set another timer until I am done renaming files
			   RenameFiles();
			   if (quitFlag == TRUE) SetTimer(hwnd,IDT_RENAME,10000, NULL);
			   else SetTimer(hwnd,IDT_RENAME,5000, NULL);
			}
			busyFlag = FALSE;
			return 0L;

      	case WM_PAINT:

			hdc=BeginPaint(hwnd,&ps);

			  if((hBitmap=LoadBitmap(hInst,"MyBmp"))!=NULL)
			  {
				if((hMemoryDC=CreateCompatibleDC(hdc))!=NULL)
				{
				  hOldBitmap=SelectObject(hMemoryDC,hBitmap);
				  if(hOldBitmap)
				  {
					GetObject(hBitmap,sizeof(BITMAP),(LPSTR)&bitmap);
					BitBlt(hdc,xborder,yborder,
							   XMAX,YMAX,
							   hMemoryDC,
							   0,0,
							   SRCCOPY);
					SelectObject(hMemoryDC,hOldBitmap);
				  }
				  DeleteDC(hMemoryDC);
				}
				DeleteObject(hBitmap);
			  }

			EndPaint(hwnd,&ps);
			return 0L;

		case WM_COMMAND:

          	wCommand = LOWORD(wParam);
          	// command message handler switch
          	// no commands yet (if ever)

          	switch(wCommand) {

				default: break;
			}
			// let DefWindowProc handle any commands that we don't
			// for any that we do handle, return 0L (LRESULT functions)
			break;

      	case WM_DESTROY:
      	    // signal timer routine that it's time to quit...
      	    quitFlag = TRUE;
            while (busyFlag == TRUE);
            KillTimer(hwnd,IDT_RENAME);
            PostQuitMessage(0);
            return 0L;
          }

     return DefWindowProc (hwnd, message, wParam, lParam) ;
}


int RenameFiles()
{

	char newfile[257];
	int status, count;


	if (quitFlag == TRUE) return -1;

	status = FileExists("2000.2000.bin");

    if (quitFlag == TRUE) return -1;
	if (status != -1) {
		count = GetLastFileCount("hgr1????.bin");
		if (quitFlag == TRUE) return -1;
		count++;
	    sprintf(newfile,"hgr1%00004d.bin",count);
	    if (quitFlag == TRUE) return -1;
	    rename("2000.2000.bin",newfile);
	    if (quitFlag == TRUE) return -1;
	}

	status = FileExists("4000.2000.bin");

    if (quitFlag == TRUE) return -1;
	if (status != -1) {
		count = GetLastFileCount("hgr2????.bin");
		if (quitFlag == TRUE) return -1;
		count++;
	    sprintf(newfile,"hgr2%00004d.bin",count);
	    if (quitFlag == TRUE) return -1;
	    rename("4000.2000.bin",newfile);
	}

    if (quitFlag == TRUE) return -1;
	return 0;

}

int GetLastFileCount(char *spec)
{
	int count, lastcount=-1;
	long findhandle;
    struct _finddata_t wild_card;
    char searchspec[128];

    if (quitFlag == TRUE) return -1;

    strcpy(searchspec, spec);
    lastcount = -1;

    if (quitFlag == TRUE) return -1;

	if((findhandle = _findfirst(searchspec, &wild_card)) > 0l)
	  {
		if (quitFlag == FALSE) {

        	count = atoi((char *)&wild_card.name[4]);
        	if (count > lastcount) {
				lastcount = count;
			}
		}
		if (quitFlag == FALSE) {

			while (_findnext(findhandle, &wild_card) == 0) {
				if (quitFlag == TRUE) break;
				count = atoi((char *)&wild_card.name[4]);
				if (count > lastcount)lastcount = count;
			}
		}
		_findclose(findhandle);
		if (quitFlag == TRUE) return -1;
	 }

	return lastcount;
}

int FileExists(char *spec)
{
	int status =-1;
	long findhandle;
    struct _finddata_t wild_card;
    char searchspec[128];

    if (quitFlag == TRUE) return -1;

    strcpy(searchspec, spec);

    if (quitFlag == FALSE) {
		if((findhandle = _findfirst(searchspec, &wild_card)) > 0l)
	  	{
		  status = 0;
		 _findclose(findhandle);
	  	}
	}

    if (quitFlag == TRUE) return -1;

	return status;
}


int ParseCommandLine(LPSTR lpszCmdParam)
{

  char DirName[257];
  int idx;

  DirName[0] = 0;
  if (NULL != lpszCmdParam) {
     if (lpszCmdParam[0] == '"')	{
       	strcpy(DirName,&lpszCmdParam[1]);
       	idx = strlen(DirName);
       	if (idx > 1 && DirName[idx-1] == '"')DirName[idx-1]=0;
     }
     else
        strcpy(DirName,lpszCmdParam);

     _chdir(DirName);
  }

  _getcwd(DirName,256);
  sprintf(MessageText,lpWelcome,DirName);
  // just use a childless messagebox
  MessageBox(NULL,MessageText,lpszTitle, MB_OK);

  return 0;

}