/* ------------------------------------------------------------------------
System       : Manx Aztec C65 Version 3.2b
               MS-DOS cross-development environment
Platform     : Apple IIe 128K PRODOS 8
Program      : ld.c
Description  : G2 Library Demo and Test Program
               List Directory

               This program lists the files in the current directory.
			   The Aztec C Shell version also provides a commandline
			   pathname option to list the files in directories other
			   than the current directory. If you are not confused
			   yet read further...

               2 versions of this program are built with the
			   MAKEFILE:

               1. A ProDOS SYS Program
               2. An Aztec C Shell PRG Program

               Separate Diskimages for each are also provided.

               Note the defines in the code below that are required
			   for compiler options to build separate programs for
			   these 2 different environments.

               I used this program as a prototype to develop filefind
			   routines for the G2 library, and left it here
			   afterwards because I think it should prove
			   informative. It's pretty well commented as is the
			   header file PRODIR.H which can be found in the INCLUDE
			   directory.

               While this program is not special and many others have
			   obviously done the same thing back in the day,
			   including Frank Gadegast (Phade Software), what I am
			   providing here is documented with technical notes
			   from Apple Computer and is free to use
			   with the usual caveats in the Licence below.

			   I mention Frank Gadegast because his work on his own
			   directory routines in Aztec C inspired my study resulting
			   in what you have here.

			   I was also inspired by Phoenyx (Mike Thomas,
			   BluPhoenyx) with his work using Phade's routines like
			   in AppleTree.

			   I recommend reading their code as well if you have
			   the time. But regardless if you take time to follow
			   the references that I have provided or not, and simply
			   follow the code that I am offering, I hope you will at
			   least have some fun with this.

Written by   : Bill Buckels
Date Written : January 2013
Revision     : 1.0 First Release
Licence      : You may use this routine for whatever you wish as long
               as you agree that Bill Buckels has no warranty or
               liability obligations whatsoever from said use.
------------------------------------------------------------------------ */

#include <stdio.h>
#include <prodir.h>

struct fileinfo pathinfo;
struct keyblockheader kbh;
struct filefind ff;


/*  ProDOS 8 Technical Reference Manual
    http://www.easy68k.com/paulrsm/6502/PDOS8TRM.HTM

    B.4.2.3 - The access Attribute

	The access attribute field determines whether the file can be
	read from, written to, deleted, or renamed. It also contains a
	bit that can be used to indicate whether a backup copy of the
	file has been made since the file's last modification.

	Figure B-14. The access Attribute Field

							   +-------- Write-Enable
							   |   +---- Read-Enable
							   |   |
	 +-------------------------------+
	 | D | RN | B | Reserved | W | R |
	 +-------------------------------+
	   |   |    |
	   |   |    +----------------------- Backup
	   |   +---------------------------- Rename-Enable
	   +-------------------------------- Destroy-Enable

	A bit set to 1 indicates that the operation is enabled; a bit
	cleared to 0 indicates that the operation is disabled. The
	reserved bits are always 0.

*/



/* create a null-terminated string of ProDOS file attributes from
   a ProDOS file access attribute bitfield */
getfattrib(access,attribs)
unsigned char access, *attribs;
{
	  attribs[0] = attribs[1] = attribs[2] = attribs[3] = attribs[4] = '-';

      if (0 != (access & 0x01)) attribs[0] = 'r';/* Read-Enable */
      if (0 != (access & 0x02)) attribs[1] = 'w';/* Write-Enable */
      if (0 != (access & 0x20)) attribs[2] = 'b';/* Backup */
      if (0 != (access & 0x40)) attribs[3] = 'n';/* Rename-Enable */
      if (0 != (access & 0x80)) attribs[4] = 'd';/* Destroy-Enable */

      attribs[5] = 0;
}

/*  ProDOS 8 Technical Reference Manual
    http://www.easy68k.com/paulrsm/6502/PDOS8TRM.HTM

Table B-1. The ProDOS File_Types

File Type      Preferred Use
 $00            Typeless file (SOS and ProDOS)
 $01            Bad block file
 $02 *          Pascal code file
 $03 *          Pascal text file
 $04            ASCII text file (SOS and ProDOS)
 $05 *          Pascal data file
 $06            General binary file (SOS and ProDOS)
 $07 *          Font file
 $08            Graphics screen file
 $09 *          Business BASIC program file
 $0A *          Business BASIC data file
 $0B *          Word Processor file
 $0C *          SOS system file
 $0D,$0E *      SOS reserved
 $0F            Directory file (SOS and ProDOS)
 $10 *          RPS data file
 $11 *          RPS index file
 $12 *          AppleFile discard file
 $13 *          AppleFile model file
 $14 *          AppleFile report format file
 $15 *          Screen Library file
 $16-$18 *      SOS reserved
 $19            AppleWorks Data Base file
 $1A            AppleWorks Word Processor file
 $1B            AppleWorks Spreadsheet file
 $1C-$EE        Reserved
 $EF            Pascal area
 $F0            ProDOS CI added command file
 $F1-$F8        ProDOS user defined files 1-8
 $F9            ProDOS reserved
 $FA            Integer BASIC program file
 $FB            Integer BASIC variable file
 $FC            Applesoft program file
 $FD            Applesoft variables file
 $FE            Relocatable code file (EDASM)
 $FF            ProDOS system file

*/

/* display some ProDOS filetypes */
char *filetypes[] = {
	"???",
	"TXT",
	"BIN",
	"DIR",
	"CMD",
	"DAT",
	"PRG",
	"BAS",
	"SYS",
	NULL};

/* index function for the string array above */
getftype(filetype)
int filetype;
{

	switch(filetype)
	{
		case 4:	  return 1;
		case 6:   return 2;
		case 15:  return 3;
		case 240: return 4;
		case 241: return 5;
		case 248: return 6;
		case 252: return 7;
		case 255: return 8;
     }
return 0;
}


foldericon()
{
	/* MouseText Characters
       Apple II Technical Notes Developer Technical Support

       http://mirrors.apple2.org.za/apple.cabi.net/FAQs.and.INFO/A2.TECH.NOTES.ETC/A2.CLASSIC.TNTS/mouse006.html

       The following displays two MouseText characters which create a folder icon...

       */

#asm

             LDA #$1B           ;MOUSETEXT ON
             JSR $FDED          ;SEND IT THROUGH THE COUT ROUTINE

             LDA #$58           ;SHOW X
             JSR $FDED          ;SEND IT THROUGH THE COUT ROUTINE

             LDA #$59           ;SHOW Y
             JSR $FDED          ;SEND IT THROUGH THE COUT ROUTINE

             LDA #$18           ;MOUSETEXT OFF
             JSR $FDED          ;SEND IT THROUGH THE COUT ROUTINE

#endasm

}

#ifdef SHELLPRG

main (argc, argv)
int argc;
char *argv[];
{

#else

main ()
{

#endif

	char pathname[64], filename[16], buf[64];
	int err = 0, fh, bytes, filelength, filetype, filecount=0, count, rowcnt;
	long size = 0L;

    crt80(); /* set to 80 column mode */
	scr_clear();
	scr_curs(0,0);

#ifdef SHELLPRG
    if (argc > 1)strcpy(pathname,argv[1]);
    else
#endif
	if (getprefix (pathname) == -1)
	{
		printf ("\nerror getting prefix!\n\n");
		err = 1;
	}

    if (err == 0) {
		if (getfinfo (pathname, &pathinfo) == -1)
		{
		   printf ("\nerror getting info for directory %s!\n\n",pathname);
		   err = 2;
		}
		else if (pathinfo.file_type != 15)
		{
		   printf ("\nerror! %s is not a directory!\n\n",pathname);
		   err = 3;

		}
		else if ((fh = open (pathname, 0, 0)) < 0)
		{

		   printf ("\nerror! can't open directory %s!\n\n",pathname);
		   err = 4;
		}
        else {

			/* 	ProDOS 8 Technical Reference Manual -
				http://www.easy68k.com/paulrsm/6502/PDOS8TRM.HTM

			B.2.4 - File Entries

			   The first entry in the key block of a directory file
			   is a header; all other entries are file entries.

			   The header entry is the same length as all other entries.

			   The first four bytes of each block used by a directory
			   file contain pointers to the preceding and succeeding
			   blocks in the directory file, respectively. Each
			   pointer is a two-byte logical block number, low byte
			   first, high byte second. The key block of a directory
			   file has no preceding block: its first pointer is
			   zero. Likewise, the last block in a directory file has
			   no successor: its second pointer is zero.

			   By the Way: All block pointers used by ProDOS have the
			   same format: low byte first, high byte second. */

			/* read and discard the first four bytes of pointer data
			   in the key block */
			bytes = read (fh, buf, 4);

			/* read the key block header entry */
			/* Thirteen fields are currently defined to be in a
			   volume directory (key block) header. */

			bytes = read (fh, &kbh.name, 39);
			if (bytes < 39)
			{
			   close (fh);
			   printf ("\nerror! can't read directory %s!\n\n",pathname);
			   err = 5;
			}
		}
	}

	if (err != 0) {
		getch();
#ifdef SHELLPRG
        exit(err);
#else
		_exit();
#endif
	}

   /* Aztec C runtime library function fixnam converts the file name into a  fully-qualified
      name  consisting of the  file name itself prefixed  by the directories that must be passed
      through to get to the file. */

    fixnam (pathname, buf);
    printf("%s\n\n",buf);

    /* read the file entries */
    count = 1;
    rowcnt = 2;

    for (;;)
    {
        bytes = read (fh, &ff.length, 39);
        if (bytes < 39) break;
        if (ff.type == 0) break;

        count++;

        if (count == 13)
        {
			/* 1. the first four bytes in each 512 byte block is pointer data.
               2. 13 file entries of 39 bytes each follow the first four bytes of pointer data.
               3. the last byte in each 512 byte block is padding.

               read and discard the last byte in the current block and the first four bytes of
               pointer data in the next block... then start reading file entries again. */
            read (fh,buf,5);
            count = 0;
        }


        filelength = (ff.length & 0xf);

        /* An entry in a directory file may be active or inactive; that is, it may
           or may not describe a file currently in the directory. If it is inactive,
           the first byte of the entry (storage_type and name_length) has the
           value zero. */

        if (filelength != 0)
        {
			filecount++;
			size += ff.size;

            strncpy(filename, ff.name, filelength);
            filename[filelength] = 0;
            filetype = getftype((int)ff.type);
            getfattrib(ff.access,(unsigned char *)&buf[0]);

            /* sub-directories display differently */
            printf("  ");

            if (ff.type == 15)foldericon();
		    else printf("  ");

			printf(" %-15s %s %s ", filename,filetypes[filetype],buf);

            if (ff.type == 15) printf("        S=%-5d\n",ff.size);
           	else printf("A=$%04x L=%-5d\n",ff.address,ff.size);

            rowcnt++;
            if (rowcnt > 20) {
			  printf("\nPress Any Key...");
			  getch();
			  scr_cr();  		/* put cursor in column 0 of next line */
			  scr_cursup();		/* goto previous line */
			  scr_ldelete();	/* delete previous line */
			  scr_cursup();		/* goto previous line */
			  scr_ldelete();	/* delete previous line */
			  rowcnt=0;
			}
        }
	}

	close (fh);
	if (filecount > 0) printf ("\n%d files using %ld bytes.", filecount, size);
	else printf("\nDirectory is empty.");
    getch();
	scr_clear();
	scr_curs(0,0);

#ifdef SHELLPRG
        exit(err);
#else
		_exit();
#endif

}